# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Set of tree views / tree-building widgets, some based on jQuery treeview
plugin.
"""


from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.predicates import adaptable
from cubicweb.utils import make_uid, json
from cubicweb_web.view import EntityView
from cubicweb_web.views import baseviews
from cubicweb_web.views.ajaxcontroller import ajaxfunc


def treecookiename(treeid):
    return str("%s-treestate" % treeid)


def _done_init(done, view, row, col):
    """handle an infinite recursion safety belt"""
    if done is None:
        done = set()
    entity = view.cw_rset.get_entity(row, col)
    if entity.eid in done:
        msg = entity._cw._("loop in %(rel)s relation (%(eid)s)") % {
            "rel": entity.cw_adapt_to("ITree").tree_relation,
            "eid": entity.eid,
        }
        return None, msg
    done.add(entity.eid)
    return done, entity


class BaseTreeView(baseviews.ListView):
    """base tree view"""

    __regid__ = "tree"
    __select__ = adaptable("ITree")
    item_vid = "treeitem"

    def call(self, done=None, **kwargs):
        if done is None:
            done = set()
        super(BaseTreeView, self).call(done=done, **kwargs)

    def cell_call(
        self, row, col=0, vid=None, done=None, maxlevel=None, klass=None, **kwargs
    ):
        assert maxlevel is None or maxlevel > 0
        done, entity = _done_init(done, self, row, col)
        if done is None:
            # entity is actually an error message
            self.w('<li class="badcontent">%s</li>', entity)
            return
        self.open_item(entity)
        entity.view(vid or self.item_vid, w=self.w, **kwargs)
        if maxlevel is not None:
            maxlevel -= 1
            if maxlevel == 0:
                self.close_item(entity)
                return
        relatedrset = entity.cw_adapt_to("ITree").children(entities=False)
        self.wview(
            self.__regid__,
            relatedrset,
            "null",
            done=done,
            maxlevel=maxlevel,
            klass=klass,
            **kwargs,
        )
        self.close_item(entity)

    def open_item(self, entity):
        self.w('<li class="%s">\n', entity.cw_etype.lower())

    def close_item(self, entity):
        self.w("</li>\n")


class TreePathView(EntityView):
    """a recursive path view"""

    __regid__ = "path"
    __select__ = adaptable("ITree")
    item_vid = "oneline"
    separator = "&#160;&gt;&#160;"

    def call(self, **kwargs):
        self.w('<div class="pathbar">')
        super(TreePathView, self).call(**kwargs)
        self.w("</div>")

    def cell_call(self, row, col=0, vid=None, done=None, **kwargs):
        done, entity = _done_init(done, self, row, col)
        if done is None:
            # entity is actually an error message
            self.w('<span class="badcontent">%s</span>', entity)
            return
        parent = entity.cw_adapt_to("ITree").parent()
        if parent:
            parent.view(self.__regid__, w=self.w, done=done)
            self.w(self.separator)
        entity.view(vid or self.item_vid, w=self.w)


class TreeComboBoxView(TreePathView):
    """display folder in edition's combobox"""

    __regid__ = "combobox"
    item_vid = "text"
    separator = " > "


# XXX rename regid to ajaxtree/foldabletree or something like that (same for
# treeitemview)
class TreeView(EntityView):
    """ajax tree view, click to expand folder"""

    __regid__ = "treeview"
    itemvid = "treeitemview"
    subvid = "oneline"
    cssclass = "treeview widget"
    title = _("tree view")

    def _init_params(self, subvid, treeid, initial_load, morekwargs):
        form = self._cw.form
        if subvid is None:
            subvid = form.pop("treesubvid", self.subvid)  # consume it
        if treeid is None:
            treeid = form.pop("treeid", None)
            if treeid is None:
                treeid = "throw_away" + make_uid("uid")
        if "morekwargs" in self._cw.form:
            ajaxargs = json.loads(form.pop("morekwargs"))
            # got unicode & python keywords must be strings
            morekwargs.update(dict((str(k), v) for k, v in ajaxargs.items()))
        return subvid, treeid

    def _init_headers(self, treeid):
        self._cw.add_css(
            ("jquery-treeview/jquery.treeview.css", "cubicweb.treeview.css")
        )
        self._cw.add_js(
            (
                "cubicweb.ajax.js",
                "cubicweb.widgets.js",
                "jquery-treeview/jquery.treeview.js",
            )
        )
        self._cw.html_headers.add_onload(
            """
jQuery("#tree-%s").treeview({toggle: toggleTree, prerendered: true});"""
            % treeid
        )

    def call(self, subvid=None, treeid=None, initial_load=True, **morekwargs):
        subvid, treeid = self._init_params(subvid, treeid, initial_load, morekwargs)
        ulid = " "
        self._init_headers(treeid)
        ulid = ' id="tree-%s"' % treeid
        self.w('<ul%s class="%s">', ulid, self.cssclass, escape=False)
        # XXX force sorting on x.sortvalue() (which return dc_title by default)
        # we need proper ITree & co specification to avoid this.
        # (pb when type ambiguity at the other side of the tree relation,
        # unability to provide generic implementation on eg Folder...)
        for i, entity in enumerate(
            sorted(self.cw_rset.entities(), key=lambda x: x.sortvalue())
        ):
            if i + 1 < len(self.cw_rset):
                morekwargs["is_last"] = False
            else:
                morekwargs["is_last"] = True
            entity.view(
                self.itemvid,
                vid=subvid,
                parentvid=self.__regid__,
                treeid=treeid,
                w=self.w,
                **morekwargs,
            )
        self.w("</ul>")

    def cell_call(self, *args, **allargs):
        """does not makes much sense until you have to invoke
        somentity.view('treeview')"""
        allargs.pop("row")
        allargs.pop("col")
        self.call(*args, **allargs)


class FileTreeView(TreeView):
    """specific version of the treeview to display file trees"""

    __regid__ = "filetree"
    cssclass = "treeview widget filetree"
    title = _("file tree view")

    def call(self, subvid=None, treeid=None, initial_load=True, **kwargs):
        super(FileTreeView, self).call(
            treeid=treeid,
            subvid="filetree-oneline",
            initial_load=initial_load,
            **kwargs,
        )


class FileItemInnerView(EntityView):
    """inner view used by the TreeItemView instead of oneline view

    This view adds an enclosing <span> with some specific CSS classes
    around the oneline view. This is needed by the jquery treeview plugin.
    """

    __regid__ = "filetree-oneline"

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        itree = entity.cw_adapt_to("ITree")
        if itree and not itree.is_leaf():
            self.w(
                '<div class="folder">%s</div>\n', entity.view("oneline"), escape=False
            )
        else:
            # XXX define specific CSS classes according to mime types
            self.w('<div class="file">%s</div>\n', entity.view("oneline"), escape=False)


class DefaultTreeViewItemView(EntityView):
    """default treeitem view for entities which don't adapt to ITree"""

    __regid__ = "treeitemview"

    def cell_call(self, row, col, vid="oneline", treeid=None, **morekwargs):
        assert treeid is not None
        itemview = self._cw.view(vid, self.cw_rset, row=row, col=col)
        last_class = morekwargs["is_last"] and ' class="last"' or ""
        self.w("<li%s>%s</li>", last_class, itemview, escape=False)


class TreeViewItemView(EntityView):
    """specific treeitem view for entities which adapt to ITree

    (each item should be expandable if it's not a tree leaf)
    """

    __regid__ = "treeitemview"
    __select__ = adaptable("ITree")
    default_branch_state_is_open = False

    def open_state(self, eeid, treeid):
        cookies = self._cw.get_cookie()
        treestate = cookies.get(treecookiename(treeid))
        if treestate:
            return str(eeid) in treestate.value.split(":")
        return self.default_branch_state_is_open

    def cell_call(
        self,
        row,
        col,
        treeid,
        vid="oneline",
        parentvid="treeview",
        is_last=False,
        **morekwargs,
    ):
        entity = self.cw_rset.get_entity(row, col)
        itree = entity.cw_adapt_to("ITree")
        liclasses = []
        if self._cw.url(includeparams=False) == entity.absolute_url():
            liclasses.append("selected")
        is_open = self.open_state(entity.eid, treeid)
        is_leaf = itree is None or itree.is_leaf()
        if is_leaf:
            if is_last:
                liclasses.append("last")
            self.w('<li class="%s">', " ".join(liclasses))
        else:
            rql = itree.children_rql() % {"x": entity.eid}
            url = xml_escape(
                self._cw.build_url(
                    "ajax",
                    rql=rql,
                    vid=parentvid,
                    pageid=self._cw.pageid,
                    treeid=treeid,
                    fname="view",
                    treesubvid=vid,
                    morekwargs=json.dumps(morekwargs),
                )
            )
            divclasses = ["hitarea"]
            if is_open:
                liclasses.append("collapsable")
                divclasses.append("collapsable-hitarea")
            else:
                liclasses.append("expandable")
                divclasses.append("expandable-hitarea")
            if is_last:
                if is_open:
                    liclasses.append("lastCollapsable")
                    divclasses.append("lastCollapsable-hitarea")
                else:
                    liclasses.append("lastExpandable")
                    divclasses.append("lastExpandable-hitarea")
            if is_open:
                self.w('<li class="%s">', " ".join(liclasses))
            else:
                self.w(
                    '<li cubicweb:loadurl="%s" class="%s">',
                    url,
                    " ".join(liclasses),
                    escape=False,
                )
            if treeid.startswith("throw_away"):
                divtail = ""
            else:
                divtail = (
                    """ onclick="asyncRemoteExec('node_clicked', '%s', '%s')" """
                    % (treeid, entity.eid)
                )
            self.w(
                '<div class="%s"%s></div>', " ".join(divclasses), divtail, escape=False
            )

            # add empty <ul> because jquery's treeview plugin checks for
            # sublists presence
            if not is_open:
                self.w('<ul class="placeholder"><li>place holder</li></ul>')
        # the local node info
        self.wview(vid, self.cw_rset, row=row, col=col, **morekwargs)
        if is_open and not is_leaf:  # => rql is defined
            self.wview(
                parentvid,
                itree.children(entities=False),
                subvid=vid,
                treeid=treeid,
                initial_load=False,
                **morekwargs,
            )
        self.w("</li>")


@ajaxfunc
def node_clicked(self, treeid, nodeeid):
    """add/remove eid in treestate cookie"""
    cookies = self._cw.get_cookie()
    statename = treecookiename(treeid)
    treestate = cookies.get(statename)
    if treestate is None:
        self._cw.set_cookie(statename, nodeeid)
    else:
        marked = set(filter(None, treestate.value.split(":")))
        if nodeeid in marked:
            marked.remove(nodeeid)
        else:
            marked.add(nodeeid)
        self._cw.set_cookie(statename, ":".join(marked))



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""an helper class to display CubicWeb schema using ureports"""


from operator import attrgetter, itemgetter

from logilab.common.ureports import Section, Title, Table, Link, Span, Text
from yams.schema import RelationDefinitionSchema
from yams.schema2dot import CARD_MAP

from cubicweb import _

TYPE_GETTER = attrgetter("type")

I18NSTRINGS = [_("read"), _("add"), _("delete"), _("update"), _("order")]


class SchemaViewer:
    """return a ureport layout for some part of a schema"""

    def __init__(self, req=None, encoding=None):
        self.req = req
        if req is not None:
            req.add_css("cubicweb.schema.css")
            if encoding is None:
                encoding = req.encoding
            self._ = req._
        else:
            encoding = "ascii"
            self._ = str
        self.encoding = encoding

    # no self.req managements

    def may_read(self, rdef, action="read"):
        """Return true if request user may read the given schema.
        Always return True when no request is provided.
        """
        if self.req is None:
            return True
        return rdef.may_have_permission(action, self.req)

    def format_eschema(self, eschema):
        text = eschema.type
        if self.req is None:
            return Text(text)
        return Link(self.req.build_url("cwetype/%s" % eschema), text)

    def format_rschema(self, rschema, label=None):
        if label is None:
            label = rschema.type
        if self.req is None:
            return Text(label)
        return Link(self.req.build_url("cwrtype/%s" % rschema), label)

    # end of no self.req managements

    def visit_schema(self, schema, display_relations=0, skiptypes=()):
        """get a layout for a whole schema"""
        title = Title(self._("Schema %s") % schema.name, klass="titleUnderline")
        layout = Section(children=(title,))
        esection = Section(
            children=(Title(self._("Entities"), klass="titleUnderline"),)
        )
        layout.append(esection)
        eschemas = [
            eschema
            for eschema in schema.entities()
            if not (eschema.final or eschema in skiptypes)
        ]
        for eschema in sorted(eschemas, key=TYPE_GETTER):
            esection.append(self.visit_entityschema(eschema, skiptypes))
        if display_relations:
            title = Title(self._("Relations"), klass="titleUnderline")
            rsection = Section(children=(title,))
            layout.append(rsection)
            relations = [
                rschema
                for rschema in sorted(schema.relations(), key=TYPE_GETTER)
                if not (rschema.final or rschema.type in skiptypes)
            ]
            keys = [(rschema.type, rschema) for rschema in relations]
            for key, rschema in sorted(keys, key=itemgetter(1)):
                relstr = self.visit_relationschema(rschema)
                rsection.append(relstr)
        return layout

    def _entity_attributes_data(self, eschema):
        _ = self._
        data = [_("attribute"), _("type"), _("default"), _("constraints")]
        attributes = sorted(eschema.attribute_definitions(), key=lambda el: el[0].type)
        for rschema, aschema in attributes:
            rdef = eschema.relation_definition(rschema)
            if not self.may_read(rdef):
                continue
            aname = rschema.type
            if aname == "eid":
                continue
            data.append("{} ({})".format(aname, _(aname)))
            data.append(_(aschema.type))
            defaultval = eschema.default(aname)
            if defaultval is not None:
                default = self.to_string(defaultval)
            elif rdef.cardinality[0] == "1":
                default = _("required field")
            else:
                default = ""
            data.append(default)
            constraints = rschema.rproperty(eschema.type, aschema.type, "constraints")
            data.append(", ".join(str(constr) for constr in constraints))
        return data

    def stereotype(self, name):
        return Span((" <<%s>>" % name,), klass="stereotype")

    def visit_entityschema(self, eschema, skiptypes=()):
        """get a layout for an entity schema"""
        etype = eschema.type
        layout = Section(children=" ", klass="clear")
        layout.append(Link(etype, "&#160;", id=etype))  # anchor
        title = self.format_eschema(eschema)
        boxchild = [Section(children=(title,), klass="title")]
        data = []
        data.append(Section(children=boxchild, klass="box"))
        data.append(Section(children="", klass="vl"))
        data.append(Section(children="", klass="hl"))
        t_vars = []
        rels = []
        first = True

        rel_defs = sorted(eschema.relation_definitions(), key=lambda el: el[0].type)

        for rschema, targetschemas, role in rel_defs:
            if rschema.type in skiptypes:
                continue
            for oeschema in sorted(targetschemas, key=TYPE_GETTER):
                rdef = rschema.role_rdef(eschema, oeschema, role)
                if not self.may_read(rdef):
                    continue
                label = rschema.type
                if role == "subject":
                    cards = rschema.relation_definition(eschema, oeschema).cardinality
                else:
                    cards = rschema.relation_definition(oeschema, eschema).cardinality
                    cards = cards[::-1]
                label = "{} {} {}".format(CARD_MAP[cards[1]], label, CARD_MAP[cards[0]])
                rlink = self.format_rschema(rschema, label)
                elink = self.format_eschema(oeschema)
                if first:
                    t_vars.append(Section(children=(elink,), klass="firstvar"))
                    rels.append(Section(children=(rlink,), klass="firstrel"))
                    first = False
                else:
                    t_vars.append(Section(children=(elink,), klass="var"))
                    rels.append(Section(children=(rlink,), klass="rel"))
        data.append(Section(children=rels, klass="rels"))
        data.append(Section(children=t_vars, klass="vars"))
        layout.append(Section(children=data, klass="entityAttributes"))
        return layout

    def visit_relationschema(self, rschema, title=True):
        """get a layout for a relation schema"""
        _ = self._
        if title:
            title = self.format_rschema(rschema)
            stereotypes = []
            if rschema.meta:
                stereotypes.append("meta")
            if rschema.symmetric:
                stereotypes.append("symmetric")
            if rschema.inlined:
                stereotypes.append("inlined")
            title = Section(children=(title,), klass="title")
            if stereotypes:
                title.append(self.stereotype(",".join(stereotypes)))
            layout = Section(children=(title,), klass="schema")
        else:
            layout = Section(klass="schema")
        data = [_("from"), _("to")]
        schema = rschema.schema
        rschema_objects = rschema.objects()
        if rschema_objects:
            # might be empty
            properties = [
                p
                for p in RelationDefinitionSchema.rproperty_defs(rschema_objects[0])
                if p not in ("cardinality", "composite", "eid")
            ]
        else:
            properties = []
        data += [_(prop) for prop in properties]
        cols = len(data)
        done = set()
        for subjtype, objtypes in sorted(rschema.associations()):
            for objtype in objtypes:
                if (subjtype, objtype) in done:
                    continue
                done.add((subjtype, objtype))
                if rschema.symmetric:
                    done.add((objtype, subjtype))
                data.append(self.format_eschema(schema[subjtype]))
                data.append(self.format_eschema(schema[objtype]))
                rdef = rschema.relation_definition(subjtype, objtype)
                for prop in properties:
                    val = getattr(rdef, prop)
                    if val is None:
                        val = ""
                    elif prop == "constraints":
                        val = ", ".join([c.expression for c in val])
                    elif isinstance(val, dict):
                        for key, value in val.items():
                            if isinstance(value, (list, tuple)):
                                val[key] = ", ".join(sorted(str(v) for v in value))
                        val = str(val)

                    elif isinstance(val, (list, tuple)):
                        val = sorted(val)
                        val = ", ".join(str(v) for v in val)
                    elif val and isinstance(val, str):
                        val = _(val)
                    else:
                        val = str(val)
                    data.append(Text(val))
        table = Table(cols=cols, rheaders=1, children=data, klass="listing")
        layout.append(Section(children=(table,), klass="relationDefinition"))
        layout.append(Section(children="", klass="clear"))
        return layout

    def to_string(self, value):
        """used to converte arbitrary values to encoded string"""
        if isinstance(value, str):
            return value.encode(self.encoding, "replace")
        return str(value)



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from urllib.parse import urljoin

from logilab.common.testlib import unittest_main

from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views import actions


class ActionsCWTC(WebCWTC):
    def test_view_action(self):
        with self.admin_access.web_request(vid="rss", rql="CWUser X") as req:
            rset = req.execute("CWUser X")
            actions = self.vreg["actions"].poss_visible_objects(req, rset=rset)
            vaction = [action for action in actions if action.__regid__ == "view"][0]
            self.assertEqual(vaction.url(), urljoin(BASE_URL, "view?rql=CWUser%20X"))

    def test_has_editable_relations(self):
        """ensure has_editable_relation predicate used by ModifyAction
        return positive score if there is only some inlined forms
        """
        # The schema only allows the anonymous user to modify his/her own
        # EmailAddress if it is set, not to create one. Since the 'anon' CWUser
        # entity is created without any associated EmailAddress entities, there
        # are no attributes nor relations that can be edited: the "modify"
        # action should not appear.
        with self.new_access("anon").web_request() as req:
            predicate = actions.has_editable_relation()
            self.assertEqual(predicate(None, req, rset=req.user.as_rset()), 0)
        # being allowed to 'add' the relation is not enough
        use_email = self.schema["use_email"].relation_definitions[
            "CWUser", "EmailAddress"
        ]
        with self.temporary_permissions((use_email, {"add": ("guests",)})):
            with self.new_access("anon").web_request() as req:
                predicate = actions.has_editable_relation()
                self.assertEqual(predicate(None, req, rset=req.user.as_rset()), 0)
        # if we also allow creating the target etype, then the "modify" action
        # should appear
        with self.temporary_permissions(
            (use_email, {"add": ("guests",)}), EmailAddress={"add": ("guests",)}
        ):
            with self.new_access("anon").web_request() as req:
                predicate = actions.has_editable_relation()
                self.assertEqual(predicate(None, req, rset=req.user.as_rset()), 1)


if __name__ == "__main__":
    unittest_main()



from cubicweb_web import formwidgets as wdgs
from cubicweb_web.views import uicfg

# fields required in the schema but automatically set by hooks. Tell about that
# to the ui
_pvdc = uicfg.autoform_field_kwargs
_pvdc.tag_attribute(
    ("File", "data_name"), {"required": False, "widget": wdgs.TextInput({"size": 45})}
)
_pvdc.tag_attribute(("File", "data_format"), {"required": False})



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""XXX rename, split, reorganize this"""

from logilab.common.registry import NoSelectableObject
from logilab.common.testlib import unittest_main

from cubicweb import Binary, UnknownProperty
from cubicweb.predicates import is_instance, specified_etype_implements, rql_condition
from cubicweb_web.action import Action
from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views import (
    primary,
    baseviews,
    tableview,
    editforms,
    management,
    actions,
    startup,
    schema,
    xbel,
    vcard,
    owl,
    treeview,
    idownloadable,
    wdoc,
    debug,
    cwuser,
    cwproperties,
    cwsources,
    xmlrss,
    rdf,
    csvexport,
    json,
    undohistory,
)

USERACTIONS = [
    actions.UserPreferencesAction,
    actions.UserInfoAction,
    actions.LogoutAction,
]
SITEACTIONS = [actions.ManageAction]
FOOTERACTIONS = [wdoc.HelpAction, wdoc.AboutAction, actions.PoweredByAction]
MANAGEACTIONS = [
    actions.SiteConfigurationAction,
    schema.ViewSchemaAction,
    cwuser.ManageUsersAction,
    cwsources.ManageSourcesAction,
    debug.SiteInfoAction,
]

if hasattr(rdf, "RDFView"):  # not available if rdflib not installed
    RDFVIEWS = [
        ("rdf", rdf.RDFView),
        ("rdf.n3", rdf.RDFN3View),
        ("rdf.turtle", rdf.RDFTurtleView),
        ("rdf.nt", rdf.RDFntriplesView),
        ("rdf.trig", rdf.RDFtrigView),
    ]
else:
    RDFVIEWS = []


class ViewSelectorCWTC(WebCWTC):
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity(
                "BlogEntry", title="une news !", content="cubicweb c'est beau"
            )
            cnx.create_entity("Bookmark", title="un signet !", path="view?vid=index")
            cnx.create_entity("EmailAddress", address="devel@logilab.fr", alias="devel")
            cnx.create_entity("Tag", name="x")
            cnx.commit()


class VRegistryTC(ViewSelectorCWTC):
    """test the view selector"""

    def _test_registered(self, registry, content):
        try:
            expected = getattr(self, "all_%s" % registry)
        except AttributeError:
            return
        if registry == "hooks":
            self.assertEqual(len(content), expected, content)
            return
        try:
            self.assertSetEqual(list(content), expected)
        except Exception:
            print(registry, sorted(expected), sorted(content))
            print("no more", [v for v in expected if v not in content])
            print("missing", [v for v in content if v not in expected])
            raise

    def setUp(self):
        super().setUp()
        assert self.vreg["views"]["propertiesform"]

    def test_possible_views_none_rset(self):
        with self.admin_access.web_request() as req:
            self.assertListEqual(
                self.pviews(req, None),
                [
                    ("cw.sources-management", cwsources.CWSourcesManagementView),
                    (
                        "cw.users-and-groups-management",
                        cwuser.UsersAndGroupsManagementView,
                    ),
                    ("gc", debug.GCView),
                    ("index", startup.IndexView),
                    ("info", debug.ProcessInformationView),
                    ("manage", startup.ManageView),
                    ("owl", owl.OWLView),
                    ("propertiesform", cwproperties.CWPropertiesForm),
                    ("registry", debug.RegistryView),
                    ("schema", schema.SchemaView),
                    ("siteinfo", debug.SiteInfoView),
                    ("systempropertiesform", cwproperties.SystemCWPropertiesForm),
                    ("undohistory", undohistory.UndoHistoryView),
                ],
            )

    def test_possible_views_noresult(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X WHERE X eid 999999")
            self.assertListEqual(
                self.pviews(req, rset),
                [
                    ("csvexport", csvexport.CSVRsetView),
                    ("ecsvexport", csvexport.CSVEntityView),
                    ("ejsonexport", json.JsonEntityView),
                    ("jsonexport", json.JsonRsetView),
                ],
            )

    def test_possible_views_one_egroup(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWGroup X WHERE X name "managers"')
            self.assertCountEqual(
                self.pviews(req, rset),
                RDFVIEWS
                + [
                    ("csvexport", csvexport.CSVRsetView),
                    ("ecsvexport", csvexport.CSVEntityView),
                    ("ejsonexport", json.JsonEntityView),
                    ("filetree", treeview.FileTreeView),
                    ("jsonexport", json.JsonRsetView),
                    ("list", baseviews.ListView),
                    ("oneline", baseviews.OneLineView),
                    ("owlabox", owl.OWLABOXView),
                    ("primary", cwuser.CWGroupPrimaryView),
                    ("rsetxml", xmlrss.XMLRsetView),
                    ("rss", xmlrss.RSSView),
                    ("sameetypelist", baseviews.SameETypeListView),
                    ("security", management.SecurityManagementView),
                    ("table", tableview.RsetTableView),
                    ("text", baseviews.TextView),
                    ("treeview", treeview.TreeView),
                    ("xbel", xbel.XbelView),
                    ("xml", xmlrss.XMLView),
                ],
            )

    def test_possible_views_multiple_egroups(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWGroup X")
            self.assertCountEqual(
                self.pviews(req, rset),
                RDFVIEWS
                + [
                    ("csvexport", csvexport.CSVRsetView),
                    ("ecsvexport", csvexport.CSVEntityView),
                    ("ejsonexport", json.JsonEntityView),
                    ("filetree", treeview.FileTreeView),
                    ("jsonexport", json.JsonRsetView),
                    ("list", baseviews.ListView),
                    ("oneline", baseviews.OneLineView),
                    ("owlabox", owl.OWLABOXView),
                    ("primary", cwuser.CWGroupPrimaryView),
                    ("rsetxml", xmlrss.XMLRsetView),
                    ("rss", xmlrss.RSSView),
                    ("sameetypelist", baseviews.SameETypeListView),
                    ("security", management.SecurityManagementView),
                    ("table", tableview.RsetTableView),
                    ("text", baseviews.TextView),
                    ("treeview", treeview.TreeView),
                    ("xbel", xbel.XbelView),
                    ("xml", xmlrss.XMLView),
                ],
            )

    def test_propertiesform_admin(self):
        assert self.vreg["views"]["propertiesform"]
        with self.admin_access.web_request() as req:
            rset1 = req.execute('CWUser X WHERE X login "admin"')
            self.assertTrue(self.vreg["views"].select("propertiesform", req, rset=None))
            self.assertTrue(
                self.vreg["views"].select("propertiesform", req, rset=rset1)
            )
            rset2 = req.execute('CWUser X WHERE X login "anon"')
            self.assertTrue(
                self.vreg["views"].select("propertiesform", req, rset=rset2)
            )

    def test_propertiesform_anon(self):
        with self.new_access("anon").web_request() as req:
            rset1 = req.execute('CWUser X WHERE X login "admin"')
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "propertiesform",
                req,
                rset=None,
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "propertiesform",
                req,
                rset=rset1,
            )
            rset2 = req.execute('CWUser X WHERE X login "anon"')
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "propertiesform",
                req,
                rset=rset2,
            )

    def test_propertiesform_jdoe(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, "jdoe")
            cnx.commit()
        with self.new_access("jdoe").web_request() as req:
            rset1 = req.execute('CWUser X WHERE X login "admin"')
            rset2 = req.execute('CWUser X WHERE X login "jdoe"')
            self.assertTrue(self.vreg["views"].select("propertiesform", req, rset=None))
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "propertiesform",
                req,
                rset=rset1,
            )
            self.assertTrue(
                self.vreg["views"].select("propertiesform", req, rset=rset2)
            )

    def test_possible_views_multiple_different_types(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X")
            self.assertCountEqual(
                self.pviews(req, rset),
                RDFVIEWS
                + [
                    ("csvexport", csvexport.CSVRsetView),
                    ("ecsvexport", csvexport.CSVEntityView),
                    ("ejsonexport", json.JsonEntityView),
                    ("filetree", treeview.FileTreeView),
                    ("jsonexport", json.JsonRsetView),
                    ("list", baseviews.ListView),
                    ("oneline", baseviews.OneLineView),
                    ("owlabox", owl.OWLABOXView),
                    ("primary", primary.PrimaryView),
                    ("rsetxml", xmlrss.XMLRsetView),
                    ("rss", xmlrss.RSSView),
                    ("security", management.SecurityManagementView),
                    ("table", tableview.RsetTableView),
                    ("text", baseviews.TextView),
                    ("treeview", treeview.TreeView),
                    ("xbel", xbel.XbelView),
                    ("xml", xmlrss.XMLView),
                ],
            )

    def test_possible_views_any_rset(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any N, X WHERE X in_group Y, Y name N")
            self.assertListEqual(
                self.pviews(req, rset),
                [
                    ("csvexport", csvexport.CSVRsetView),
                    ("jsonexport", json.JsonRsetView),
                    ("rsetxml", xmlrss.XMLRsetView),
                    ("table", tableview.RsetTableView),
                ],
            )

    def test_possible_views_multiple_eusers(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X")
            self.assertCountEqual(
                self.pviews(req, rset),
                RDFVIEWS
                + [
                    ("csvexport", csvexport.CSVRsetView),
                    ("ecsvexport", csvexport.CSVEntityView),
                    ("ejsonexport", json.JsonEntityView),
                    ("filetree", treeview.FileTreeView),
                    ("foaf", cwuser.FoafView),
                    ("jsonexport", json.JsonRsetView),
                    ("list", baseviews.ListView),
                    ("oneline", baseviews.OneLineView),
                    ("owlabox", owl.OWLABOXView),
                    ("primary", primary.PrimaryView),
                    ("rsetxml", xmlrss.XMLRsetView),
                    ("rss", xmlrss.RSSView),
                    ("sameetypelist", baseviews.SameETypeListView),
                    ("security", management.SecurityManagementView),
                    ("table", tableview.RsetTableView),
                    ("text", baseviews.TextView),
                    ("treeview", treeview.TreeView),
                    ("vcard", vcard.VCardCWUserView),
                    ("xbel", xbel.XbelView),
                    ("xml", xmlrss.XMLView),
                ],
            )

    def test_possible_actions_none_rset(self):
        with self.admin_access.web_request() as req:
            self.assertDictEqual(
                self.pactionsdict(req, None, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "manage": MANAGEACTIONS,
                    "footer": FOOTERACTIONS,
                },
            )

    def test_possible_actions_no_entity(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X WHERE X eid 999999")
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "manage": MANAGEACTIONS,
                    "footer": FOOTERACTIONS,
                },
            )

    def test_possible_actions_same_type_entities(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWGroup X")
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "manage": MANAGEACTIONS,
                    "footer": FOOTERACTIONS,
                    "mainactions": [actions.MultipleEditAction],
                    "moreactions": [actions.DeleteAction, actions.AddNewAction],
                },
            )

    def test_possible_actions_different_types_entities(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X")
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "manage": MANAGEACTIONS,
                    "footer": FOOTERACTIONS,
                    "moreactions": [actions.DeleteAction],
                },
            )

    def test_possible_actions_final_entities(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any N, X WHERE X in_group Y, Y name N")
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "manage": MANAGEACTIONS,
                    "footer": FOOTERACTIONS,
                },
            )

    def test_possible_actions_eetype_cwuser_entity(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWEType X WHERE X name "CWUser"')
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "manage": MANAGEACTIONS,
                    "footer": FOOTERACTIONS,
                    "mainactions": [actions.ModifyAction, actions.ViewSameCWEType],
                    "moreactions": [
                        actions.ManagePermissionsAction,
                        actions.AddRelatedActions,
                        actions.DeleteAction,
                        actions.CopyAction,
                    ],
                },
            )

    def test_select_creation_form(self):
        rset = None
        with self.admin_access.web_request() as req:
            # creation form
            req.form["etype"] = "CWGroup"
            self.assertIsInstance(
                self.vreg["views"].select("creation", req, rset=rset),
                editforms.CreationFormView,
            )

        with self.admin_access.web_request() as req:
            # custom creation form
            class CWUserCreationForm(editforms.CreationFormView):
                __select__ = specified_etype_implements("CWUser")

            self.vreg._loadedmods[__name__] = {}
            self.vreg.register(CWUserCreationForm)
            req.form["etype"] = "CWUser"

            self.assertIsInstance(
                self.vreg["views"].select("creation", req, rset=rset),
                CWUserCreationForm,
            )

    def test_select_view(self):
        # no entity
        rset = None
        with self.admin_access.web_request() as req:
            self.assertIsInstance(
                self.vreg["views"].select("index", req, rset=rset), startup.IndexView
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "primary", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "table", req, rset=rset
            )

            # no entity
            rset = req.execute("Any X WHERE X eid 999999")
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "index", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "creation",
                req,
                rset=rset,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "primary", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "table", req, rset=rset
            )
            # one entity
            rset = req.execute('CWGroup X WHERE X name "managers"')
            self.assertIsInstance(
                self.vreg["views"].select("primary", req, rset=rset),
                primary.PrimaryView,
            )
            self.assertIsInstance(
                self.vreg["views"].select("list", req, rset=rset), baseviews.ListView
            )
            self.assertIsInstance(
                self.vreg["views"].select("edition", req, rset=rset),
                editforms.EditionFormView,
            )
            self.assertIsInstance(
                self.vreg["views"].select("table", req, rset=rset),
                tableview.RsetTableView,
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "creation",
                req,
                rset=rset,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "index", req, rset=rset
            )
            # list of entities of the same type
            rset = req.execute("CWGroup X")
            self.assertIsInstance(
                self.vreg["views"].select("primary", req, rset=rset),
                primary.PrimaryView,
            )
            self.assertIsInstance(
                self.vreg["views"].select("list", req, rset=rset), baseviews.ListView
            )
            self.assertIsInstance(
                self.vreg["views"].select("table", req, rset=rset),
                tableview.RsetTableView,
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "creation",
                req,
                rset=rset,
            )
            # list of entities of different types
            rset = req.execute("Any X")
            self.assertIsInstance(
                self.vreg["views"].select("primary", req, rset=rset),
                primary.PrimaryView,
            )
            self.assertIsInstance(
                self.vreg["views"].select("list", req, rset=rset), baseviews.ListView
            )
            self.assertIsInstance(
                self.vreg["views"].select("table", req, rset=rset),
                tableview.RsetTableView,
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "creation",
                req,
                rset=rset,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "index", req, rset=rset
            )
            # whatever
            rset = req.execute("Any N, X WHERE X in_group Y, Y name N")
            self.assertIsInstance(
                self.vreg["views"].select("table", req, rset=rset),
                tableview.RsetTableView,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "index", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "creation",
                req,
                rset=rset,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "primary", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "list", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "edition", req, rset=rset
            )
            # mixed query
            rset = req.execute("Any U,G WHERE U is CWUser, G is CWGroup")
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "edition", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject,
                self.vreg["views"].select,
                "creation",
                req,
                rset=rset,
            )
            self.assertIsInstance(
                self.vreg["views"].select("table", req, rset=rset),
                tableview.RsetTableView,
            )

    def test_interface_selector(self):
        with self.admin_access.web_request() as req:
            req.create_entity("File", data_name="bim.png", data=Binary(b"bim"))
            # image primary view priority
            rset = req.execute('File X WHERE X data_name "bim.png"')
            self.assertIsInstance(
                self.vreg["views"].select("primary", req, rset=rset),
                idownloadable.IDownloadablePrimaryView,
            )

    def test_score_entity_selector(self):
        with self.admin_access.web_request() as req:
            req.create_entity("File", data_name="bim.png", data=Binary(b"bim"))
            # image/ehtml primary view priority
            rset = req.execute('File X WHERE X data_name "bim.png"')
            self.assertIsInstance(
                self.vreg["views"].select("image", req, rset=rset),
                idownloadable.ImageView,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "ehtml", req, rset=rset
            )

            req.create_entity(
                "File", data_name="bim.html", data=Binary(b"<html>bam</html")
            )
            # image/ehtml primary view priority
            rset = req.execute('File X WHERE X data_name "bim.html"')
            self.assertIsInstance(
                self.vreg["views"].select("ehtml", req, rset=rset),
                idownloadable.EHTMLView,
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "image", req, rset=rset
            )

            req.create_entity("File", data_name="bim.txt", data=Binary(b"boum"))
            # image/ehtml primary view priority
            rset = req.execute('File X WHERE X data_name "bim.txt"')
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "image", req, rset=rset
            )
            self.assertRaises(
                NoSelectableObject, self.vreg["views"].select, "ehtml", req, rset=rset
            )

    def _test_view(self, vid, rql, args):
        with self.admin_access.web_request() as req:
            if rql is None:
                rset = None
            else:
                rset = req.execute(rql)
            try:
                obj = self.vreg["views"].select(vid, req, rset=rset, **args)
                return obj.render(**args)
            except Exception:
                print(vid, rset, args)
                raise

    def test_form(self):
        for vid, rql, args in (
            # ('creation', 'Any X WHERE X eid 999999', {}),
            ("edition", 'CWGroup X WHERE X name "managers"', {}),
            ("copy", 'CWGroup X WHERE X name "managers"', {}),
            ("muledit", "CWGroup X", {}),
            # ('muledit', 'Any X', {}),
        ):
            self._test_view(vid, rql, args)

    def test_properties(self):
        self.assertEqual(
            sorted(
                k
                for k in self.vreg["propertydefs"]
                if k.startswith("ctxcomponents.edit_box")
            ),
            [
                "ctxcomponents.edit_box.context",
                "ctxcomponents.edit_box.order",
                "ctxcomponents.edit_box.visible",
            ],
        )
        self.assertEqual(
            [
                k
                for k in self.vreg["propertyvalues"]
                if not k.startswith("system.version")
            ],
            [],
        )
        self.assertEqual(
            self.vreg.property_value("ctxcomponents.edit_box.visible"), True
        )
        self.assertEqual(self.vreg.property_value("ctxcomponents.edit_box.order"), 2)
        self.assertEqual(
            self.vreg.property_value("ctxcomponents.possible_views_box.visible"), False
        )
        self.assertEqual(
            self.vreg.property_value("ctxcomponents.possible_views_box.order"), 10
        )
        self.assertRaises(
            UnknownProperty, self.vreg.property_value, "ctxcomponents.actions_box"
        )


class CWETypeRQLAction(Action):
    __regid__ = "testaction"
    __select__ = is_instance("CWEType") & rql_condition('X name "CWEType"')
    title = "bla"


class RQLActionTC(ViewSelectorCWTC):
    def setUp(self):
        super().setUp()
        self.vreg._loadedmods[__name__] = {}
        self.vreg.register(CWETypeRQLAction)
        actionsreg = self.vreg["actions"]
        actionsreg["testaction"][0].__registered__(actionsreg)

    def tearDown(self):
        super().tearDown()
        del self.vreg["actions"]["testaction"]

    def test(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWEType X WHERE X name "CWEType"')
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "footer": FOOTERACTIONS,
                    "manage": MANAGEACTIONS,
                    "mainactions": [actions.ModifyAction, actions.ViewSameCWEType],
                    "moreactions": [
                        actions.ManagePermissionsAction,
                        actions.AddRelatedActions,
                        actions.DeleteAction,
                        actions.CopyAction,
                        CWETypeRQLAction,
                    ],
                },
            )
            rset = req.execute('CWEType X WHERE X name "CWRType"')
            self.assertDictEqual(
                self.pactionsdict(req, rset, skipcategories=()),
                {
                    "useractions": USERACTIONS,
                    "siteactions": SITEACTIONS,
                    "footer": FOOTERACTIONS,
                    "manage": MANAGEACTIONS,
                    "mainactions": [actions.ModifyAction, actions.ViewSameCWEType],
                    "moreactions": [
                        actions.ManagePermissionsAction,
                        actions.AddRelatedActions,
                        actions.DeleteAction,
                        actions.CopyAction,
                    ],
                },
            )


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""management and error screens"""

from time import strftime, localtime

from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.predicates import none_rset, match_user_groups
from cubicweb_web.view import StartupView
from cubicweb_web.views import actions, tabs


def dict_to_html(w, dict):
    # XHTML doesn't allow emtpy <ul> nodes
    if dict:
        w("<ul>")
        for key in sorted(dict):
            w("<li><span>%s</span>: <span>%s</span></li>", str(key), repr(dict[key]))
        w("</ul>")


class SiteInfoAction(actions.ManagersAction):
    __regid__ = "siteinfo"
    __select__ = match_user_groups("users", "managers")
    title = _("Site information")
    category = "manage"
    order = 1000


class SiteInfoView(tabs.TabsMixin, StartupView):
    __regid__ = "siteinfo"
    title = _("Site information")
    tabs = [_("info"), _("registry"), _("gc")]
    default_tab = "info"

    def call(self, **kwargs):
        """The default view representing the instance's management"""
        self.w("<h1>%s</h1>", self._cw._(self.title))
        self.render_tabs(self.tabs, self.default_tab)


class ProcessInformationView(StartupView):
    """display various web server /repository information"""

    __regid__ = "info"
    __select__ = none_rset() & match_user_groups("managers", "users")

    title = _("server information")
    cache_max_age = 0

    def call(self, **kwargs):
        req = self._cw
        dtformat = req.property_value("ui.datetime-format")
        _ = req._
        repo = req.cnx.repo
        # generic instance information
        self.w("<h2>%s</h2>", _("Instance"))
        pyvalue = (
            (_("config type"), self._cw.vreg.config.name),
            (_("config mode"), self._cw.vreg.config.mode),
            (_("instance home"), self._cw.vreg.config.apphome),
        )
        self.wview("pyvaltable", pyvalue=pyvalue, header_column_idx=0)
        vcconf = repo.get_versions()
        self.w("<h3>%s</h3>", _("versions configuration"))
        missing = _("no version information")
        pyvalue = [("CubicWeb", vcconf.get("cubicweb", missing))]
        pyvalue += [
            (cube, vcconf.get(cube, missing))
            for cube in sorted(self._cw.vreg.config.cubes())
        ]
        self.wview("pyvaltable", pyvalue=pyvalue, header_column_idx=0)
        # repository information
        self.w("<h2>%s</h2>", _("Repository"))
        self.w("<h3>%s</h3>", _("resources usage"))
        stats = self._cw.call_service("repo_stats")
        stats["threads"] = ", ".join(sorted(stats["threads"]))
        for k in stats:
            if k == "type_cache_size":
                continue
            if k.endswith("_cache_size"):
                stats[k] = "{} / {}".format(stats[k]["size"], stats[k]["maxsize"])

        def format_stat(sname, sval):
            return "{} {}".format(
                xml_escape(str(sval)),
                sname.endswith("percent") and "%" or "",
            )

        pyvalue = [
            (sname, format_stat(sname, sval)) for sname, sval in sorted(stats.items())
        ]
        self.wview("pyvaltable", pyvalue=pyvalue, header_column_idx=0)
        # web server information
        self.w("<h2>%s</h2>", _("Web server"))
        pyvalue = (
            (_("base url"), req.base_url()),
            (_("data directory url"), req.datadir_url),
        )
        self.wview("pyvaltable", pyvalue=pyvalue, header_column_idx=0)
        from cubicweb_web.application import SESSION_MANAGER

        if SESSION_MANAGER is not None and req.user.is_in_group("managers"):
            sessions = SESSION_MANAGER.current_sessions()
            self.w("<h3>%s</h3>", _("opened web sessions"))
            if sessions:
                self.w("<ul>")
                for session in sessions:
                    last_usage_time = session.mtime
                    self.w(
                        "<li>%s (%s: %s)<br/>",
                        session.sessionid,
                        _("last usage"),
                        strftime(dtformat, localtime(last_usage_time)),
                    )
                    dict_to_html(self.w, session.data)
                    self.w("</li>")
                self.w("</ul>")
            else:
                self.w("<p>%s</p>", _("no web sessions found"))


class RegistryView(StartupView):
    """display vregistry content"""

    __regid__ = "registry"
    __select__ = StartupView.__select__ & match_user_groups("managers")
    title = _("registry")
    cache_max_age = 0

    def call(self, **kwargs):
        self.w("<h2>%s</h2>", self._cw._("Registry's content"))
        keys = sorted(self._cw.vreg)
        self.w(
            "<p>%s</p>\n",
            " - ".join(f'<a href="#{key}">{key}</a>' for key in keys),
            escape=False,
        )
        for key in keys:
            if key in ("boxes", "contentnavigation"):  # those are bw compat registries
                continue
            self.w('<h3 id="%s">%s</h3>', key, key)
            if self._cw.vreg[key]:
                values = sorted(self._cw.vreg[key].items())
                self.wview(
                    "pyvaltable",
                    pyvalue=[(key, xml_escape(repr(val))) for key, val in values],
                )
            else:
                self.w("<p>Empty</p>\n")


class GCView(StartupView):
    """display garbage collector information"""

    __regid__ = "gc"
    __select__ = StartupView.__select__ & match_user_groups("managers")
    title = _("memory leak debugging")
    cache_max_age = 0

    def call(self, **kwargs):
        stats = self._cw.call_service("repo_gc_stats")
        self.w("<h2>%s</h2>", _("Garbage collection information"))
        self.w("<h3>%s</h3>", self._cw._("Looked up classes"))
        self.wview("pyvaltable", pyvalue=stats["lookupclasses"])
        self.w("<h3>%s</h3>", self._cw._("Most referenced classes"))
        self.wview("pyvaltable", pyvalue=stats["referenced"])
        if stats["unreachable"]:
            self.w("<h3>%s</h3>", self._cw._("Unreachable objects"))
            values = [xml_escape(val) for val in stats["unreachable"]]
            self.wview("pyvallist", pyvalue=values)



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Bases HTML components:

* the rql input form
* the logged user link
"""

from logilab.common.deprecation import class_renamed
from logilab.common.registry import yes

from cubicweb import _
from cubicweb.predicates import (
    match_context,
    configuration_values,
    anonymous_user,
    authenticated_user,
)
from cubicweb.uilib import toggle_action
from cubicweb.utils import wrap_on_write
from cubicweb_web import component
from cubicweb_web.htmlwidgets import MenuWidget, PopupBoxMenu

VISIBLE_PROP_DEF = {
    _("visible"): dict(
        type="Boolean", default=True, help=_("display the component or not")
    ),
}


class RQLInputForm(component.Component):
    """build the rql input form, usually displayed in the header"""

    __regid__ = "rqlinput"
    cw_property_defs = VISIBLE_PROP_DEF
    visible = False

    def call(self, view=None):
        req = self._cw
        if hasattr(view, "filter_box_context_info"):
            rset = view.filter_box_context_info()[0]
        else:
            rset = self.cw_rset
        # display multilines query as one line
        rql = rset is not None and rset.printable_rql() or req.form.get("rql", "")
        rql = rql.replace("\n", " ")
        rql_suggest_ajax_func = self._cw.vreg["ajax-func"].select_or_none(
            "rql_suggest", self._cw
        )
        if rql_suggest_ajax_func is not None:
            # enable autocomplete feature only if the rql
            # suggestions builder is available
            self._cw.add_css("jquery.ui.css")
            self._cw.add_js(("cubicweb.ajax.js", "jquery.ui.js"))
            self._cw.add_onload(
                '$("#rql").autocomplete({source: "%s"});'
                % (req.build_url("ajax", fname="rql_suggest"))
            )
        self.w(
            """<div id="rqlinput" class="%s"><form action="%s"><fieldset>
<input type="text" id="rql" name="rql" value="%s"  title="%s" accesskey="q" class="searchField" />
""",
            not self.cw_propval("visible") and "hidden" or "",
            req.build_url("view"),
            rql,
            req._("full text or RQL query"),
        )
        if req.search_state[0] != "normal":
            self.w(
                '<input type="hidden" name="__mode" value="%s"/>',
                ":".join(req.search_state[1]),
            )
        self.w("</fieldset></form></div>")


class HeaderComponent(
    component.CtxComponent
):  # XXX rename properly along with related context
    """if the user is the anonymous user, build a link to login else display a menu
    with user'action (preference, logout, etc...)
    """

    __abstract__ = True
    cw_property_defs = component.override_ctx(
        component.CtxComponent,
        vocabulary=[
            "header-center",
            "header-left",
            "header-right",
        ],
    )
    # don't want user to hide this component using an cwproperty
    site_wide = True
    context = _("header-center")


class ApplLogo(HeaderComponent):
    """build the instance logo, usually displayed in the header"""

    __regid__ = "logo"
    __select__ = yes()  # no need for a cnx
    order = -1
    context = _("header-left")

    def render(self, w):
        w('<a id="logo" href="%s"></a>' % self._cw.base_url())


class ApplicationName(HeaderComponent):
    """display the instance name"""

    __regid__ = "appliname"

    # XXX support kwargs for compat with other components which gets the view as
    # argument
    def render(self, w, **kwargs):
        title = self._cw.property_value("ui.site-title")
        if title:
            w(
                '<span id="appliName"><a href="%s">%s</a></span>',
                self._cw.base_url(),
                title,
            )


class CookieLoginComponent(HeaderComponent):
    __regid__ = "anonuserlink"
    __select__ = (
        HeaderComponent.__select__
        & anonymous_user()
        & configuration_values("auth-mode", "cookie")
    )
    context = "header-right"
    loginboxid = "popupLoginBox"
    _html = """<a class="logout icon-login" title="%s" href="javascript:
cw.htmlhelpers.popupLoginBox('%s', '__login');">%s</a>"""

    def render(self, w):
        # XXX bw compat, though should warn about subclasses redefining call
        self.w = w
        self.call()

    def call(self):
        self._cw.add_css("cubicweb.pictograms.css")
        self.w(
            self._html,
            self._cw._("login / password"),
            self.loginboxid,
            self._cw._("i18n_login_popup"),
        )
        self._cw.view(
            "logform",
            rset=self.cw_rset,
            id=self.loginboxid,
            klass="%s hidden" % self.loginboxid,
            title=False,
            showmessage=False,
            w=self.w,
        )


class HTTPLoginComponent(CookieLoginComponent):
    __select__ = (
        HeaderComponent.__select__
        & anonymous_user()
        & configuration_values("auth-mode", "http")
    )

    def render(self, w):
        # this redirects to the 'login' controller which in turn
        # will raise a 401/Unauthorized
        req = self._cw
        w(
            '[<a class="logout" title="%s" href="%s">%s</a>]'
            % (req._("login / password"), req.build_url("login"), req._("login"))
        )


_UserLink = class_renamed("_UserLink", HeaderComponent)
AnonUserLink = class_renamed("AnonUserLink", CookieLoginComponent)
AnonUserLink.__abstract__ = True
AnonUserLink.__select__ &= yes(1)


class AnonUserStatusLink(HeaderComponent):
    __regid__ = "userstatus"
    __select__ = anonymous_user()
    context = _("header-right")
    order = HeaderComponent.order - 10

    def render(self, w):
        pass


class AuthenticatedUserStatus(AnonUserStatusLink):
    __select__ = authenticated_user()

    def render(self, w):
        # display useractions and siteactions
        self._cw.add_css("cubicweb.pictograms.css")
        actions = self._cw.vreg["actions"].possible_actions(
            self._cw, rset=self.cw_rset, view=self.cw_extra_kwargs["view"]
        )
        box = MenuWidget("", "userActionsBox", _class="", islist=False)
        menu = PopupBoxMenu(self._cw.user.login, isitem=False, link_class="icon-user")
        box.append(menu)
        for action in actions.get("useractions", ()):
            menu.append(self.action_link(action))
        if actions.get("useractions") and actions.get("siteactions"):
            menu.append(self.separator())
        for action in actions.get("siteactions", ()):
            menu.append(self.action_link(action))
        box.render(w=w)


class ApplicationMessage(component.Component):
    """display messages given using the __message/_cwmsgid parameter into a
    special div section
    """

    __select__ = yes()
    __regid__ = "applmessages"
    # don't want user to hide this component using a cwproperty
    cw_property_defs = {}

    def call(self, msg=None):
        if msg is None:
            msg = self._cw.message  # XXX don't call self._cw.message twice
        self.w(
            '<div id="appMsg" onclick="%s" class="%s">\n',
            toggle_action("appMsg"),
            (msg and " " or "hidden"),
        )
        self.w('<div class="message" id="%s">', self.domid)
        self.w(str(msg), escape=False)
        self.w("</div>")
        self.w("</div>")


# contextual components ########################################################


class MetaDataComponent(component.EntityCtxComponent):
    __regid__ = "metadata"
    context = "navbottom"
    order = 1

    def render_body(self, w):
        self.entity.view("metadata", w=w)


class SectionLayout(component.Layout):
    __select__ = match_context(
        "navtop", "navbottom", "navcontenttop", "navcontentbottom"
    )
    cssclass = "section"

    def render(self, w):
        if self.init_rendering():
            view = self.cw_extra_kwargs["view"]
            w(
                '<div class="%s %s" id="%s">'
                % (self.cssclass, view.cssclass, view.domid)
            )
            with wrap_on_write(w, "<h4>") as wow:
                view.render_title(wow)
            view.render_body(w)
            w("</div>\n")



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common import tempattr
from logilab.common.testlib import unittest_main, mock_object

from cubicweb.devtools import BASE_URL
from cubicweb.predicates import is_instance
from cubicweb.schema import RRQLExpression

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.formwidgets import AutoCompletionWidget
from cubicweb_web.views import uicfg
from cubicweb_web.views.editforms import DeleteConfFormView

AFFK = uicfg.autoform_field_kwargs
AFS = uicfg.autoform_section


def rbc(entity, formtype, section):
    if section in ("attributes", "metadata", "hidden"):
        permission = "update"
    else:
        permission = "add"
    return [
        (rschema.type, x)
        for rschema, tschemas, x in AFS.relations_by_section(
            entity, formtype, section, permission
        )
    ]


class AutomaticEntityFormCWTC(WebCWTC):
    def test_custom_widget(self):
        with self.admin_access.web_request() as req:
            AFFK.tag_subject_of(
                ("CWUser", "login", "*"),
                {"widget": AutoCompletionWidget(autocomplete_initfunc="get_logins")},
            )
            form = self.vreg["forms"].select("edition", req, entity=req.user)
            field = form.field_by_name("login", "subject")
            self.assertIsInstance(field.widget, AutoCompletionWidget)
            AFFK.del_rtag("CWUser", "login", "*", "subject")

    def test_cwuser_relations_by_category(self):
        with self.admin_access.web_request() as req:
            e = self.vreg["etypes"].etype_class("CWUser")(req)
            # see custom configuration in views.cwuser
            self.assertEqual(
                rbc(e, "main", "attributes"),
                [
                    ("login", "subject"),
                    ("upassword", "subject"),
                    ("firstname", "subject"),
                    ("surname", "subject"),
                    ("in_group", "subject"),
                ],
            )
            self.assertEqual(
                rbc(e, "muledit", "attributes"),
                [
                    ("login", "subject"),
                    ("upassword", "subject"),
                    ("in_group", "subject"),
                ],
            )
            self.assertCountEqual(
                rbc(e, "main", "metadata"),
                [
                    ("last_login_time", "subject"),
                    ("cw_source", "subject"),
                    ("creation_date", "subject"),
                    ("modification_date", "subject"),
                    ("created_by", "subject"),
                    ("owned_by", "subject"),
                    ("bookmarked_by", "object"),
                ],
            )
            # XXX skip 'tags' relation here and in the hidden category because
            # of some test interdependancy when pytest is launched on whole cw
            # (appears here while expected in hidden
            self.assertCountEqual(
                [x for x in rbc(e, "main", "relations") if x != ("tags", "object")],
                [
                    ("connait", "subject"),
                    ("custom_workflow", "subject"),
                    ("primary_email", "subject"),
                    ("evaluee", "subject"),
                    ("checked_by", "object"),
                ],
            )
            self.assertListEqual(
                rbc(e, "main", "inlined"),
                [
                    ("use_email", "subject"),
                ],
            )
            # owned_by is defined both as subject and object relations on CWUser
            self.assertListEqual(
                sorted(x for x in rbc(e, "main", "hidden") if x != ("tags", "object")),
                sorted(
                    [
                        ("for_user", "object"),
                        ("created_by", "object"),
                        ("wf_info_for", "object"),
                        ("owned_by", "object"),
                    ]
                ),
            )

    def test_inlined_view(self):
        self.assertIn(
            "main_inlined",
            AFS.etype_get("CWUser", "use_email", "subject", "EmailAddress"),
        )
        self.assertNotIn(
            "main_inlined",
            AFS.etype_get("CWUser", "primary_email", "subject", "EmailAddress"),
        )
        self.assertIn(
            "main_relations",
            AFS.etype_get("CWUser", "primary_email", "subject", "EmailAddress"),
        )

    def test_personne_relations_by_category(self):
        with self.admin_access.web_request() as req:
            e = self.vreg["etypes"].etype_class("Personne")(req)
            self.assertListEqual(
                rbc(e, "main", "attributes"),
                [
                    ("nom", "subject"),
                    ("prenom", "subject"),
                    ("type", "subject"),
                    ("sexe", "subject"),
                    ("promo", "subject"),
                    ("titre", "subject"),
                    ("ass", "subject"),
                    ("web", "subject"),
                    ("tel", "subject"),
                    ("fax", "subject"),
                    ("datenaiss", "subject"),
                    ("tzdatenaiss", "subject"),
                    ("test", "subject"),
                    ("description", "subject"),
                    ("salary", "subject"),
                ],
            )
            self.assertListEqual(
                rbc(e, "muledit", "attributes"),
                [
                    ("nom", "subject"),
                ],
            )
            self.assertCountEqual(
                rbc(e, "main", "metadata"),
                [
                    ("cw_source", "subject"),
                    ("creation_date", "subject"),
                    ("modification_date", "subject"),
                    ("created_by", "subject"),
                    ("owned_by", "subject"),
                ],
            )
            self.assertCountEqual(
                rbc(e, "main", "relations"),
                [
                    ("travaille", "subject"),
                    ("manager", "object"),
                    ("tags", "subject"),
                    ("evaluee", "subject"),
                    ("actionnaire", "subject"),
                    ("dirige", "subject"),
                    ("associe", "subject"),
                    ("connait", "subject"),
                    ("tags", "object"),
                    ("contrat_exclusif", "object"),
                    ("ecrit_par", "object"),
                    ("evaluee", "object"),
                    ("associe", "object"),
                    ("fabrique_par", "object"),
                ],
            )
            self.assertListEqual(rbc(e, "main", "hidden"), [])

    def test_edition_form(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X LIMIT 1")
            form = self.vreg["forms"].select("edition", req, rset=rset, row=0, col=0)
            # should be also selectable by specifying entity
            self.vreg["forms"].select("edition", req, entity=rset.get_entity(0, 0))
            self.assertFalse(any(f for f in form.fields if f is None))

    def test_edition_form_with_action(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X LIMIT 1")
            form = self.vreg["forms"].select(
                "edition", req, rset=rset, row=0, col=0, action="my_custom_action"
            )
            self.assertEqual(form.form_action(), "my_custom_action")

    def test_attribute_add_permissions(self):
        # https://www.cubicweb.org/ticket/4342844
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, "toto")
            cnx.commit()
        with self.new_access("toto").web_request() as req:
            e = self.vreg["etypes"].etype_class("Personne")(req)
            cform = self.vreg["forms"].select("edition", req, entity=e)
            self.assertIn(
                "sexe", [rschema.type for rschema, _ in cform.editable_attributes()]
            )
            with self.new_access("toto").repo_cnx() as cnx:
                person_eid = cnx.create_entity("Personne", nom="Robert").eid
                cnx.commit()
            person = req.entity_from_eid(person_eid)
            mform = self.vreg["forms"].select("edition", req, entity=person)
            self.assertNotIn(
                "sexe", [rschema.type for rschema, _ in mform.editable_attributes()]
            )

    def test_editable_attributes(self):
        with self.admin_access.web_request() as req:
            entity = self.vreg["etypes"].etype_class("Personne")(req)
            form = self.vreg["forms"].select("edition", req, entity=entity)
            expected = [
                ("nom", "subject"),
                ("prenom", "subject"),
                ("type", "subject"),
                ("sexe", "subject"),
                ("promo", "subject"),
                ("titre", "subject"),
                ("ass", "subject"),
                ("web", "subject"),
                ("tel", "subject"),
                ("fax", "subject"),
                ("datenaiss", "subject"),
                ("tzdatenaiss", "subject"),
                ("test", "subject"),
                ("description", "subject"),
                ("salary", "subject"),
            ]
            self.assertEqual(
                [(rschema.type, role) for rschema, role in form.editable_attributes()],
                expected,
            )

            # now with display_fields attribute set
            form.display_fields = [("nom", "subject")]
            expected = [("nom", "subject")]
            self.assertEqual(
                [(rschema.type, role) for rschema, role in form.editable_attributes()],
                expected,
            )

    def test_inlined_relations(self):
        with self.admin_access.web_request() as req:
            with self.temporary_permissions(EmailAddress={"add": ()}):
                autoform = self.vreg["forms"].select("edition", req, entity=req.user)
                self.assertEqual(list(autoform.inlined_form_views()), [])

    def test_inlined_form_views(self):
        # when some relation has + cardinality, and some already linked entities which are not
        # updatable, a link to optionally add a new sub-entity should be displayed, not a sub-form
        # forcing creation of a sub-entity
        from cubicweb_web.views import autoform

        with self.admin_access.web_request() as req:
            req.create_entity(
                "EmailAddress",
                address="admin@cubicweb.org",
                reverse_use_email=req.user.eid,
            )
            use_email_schema = self.vreg.schema["CWUser"].relation_definition(
                "use_email"
            )
            with tempattr(use_email_schema, "cardinality", "+1"):
                with self.temporary_permissions(EmailAddress={"update": ()}):
                    form = self.vreg["forms"].select("edition", req, entity=req.user)
                    formviews = list(form.inlined_form_views())
                    self.assertEqual(len(formviews), 1, formviews)
                    self.assertIsInstance(formviews[0], autoform.InlineAddNewLinkView)
            # though do not introduce regression on entity creation with 1 cardinality relation
            with tempattr(use_email_schema, "cardinality", "11"):
                user = self.vreg["etypes"].etype_class("CWUser")(req)
                form = self.vreg["forms"].select("edition", req, entity=user)
                formviews = list(form.inlined_form_views())
                self.assertEqual(len(formviews), 1, formviews)
                self.assertIsInstance(
                    formviews[0], autoform.InlineEntityCreationFormView
                )

    def test_check_inlined_rdef_permissions(self):
        # try to check permissions when creating an entity ('user' below is a
        # fresh entity without an eid)
        with self.admin_access.web_request() as req:
            ttype = "EmailAddress"
            rschema = self.schema["use_email"]
            rdef = rschema.relation_definitions[("CWUser", ttype)]
            tschema = self.schema[ttype]
            role = "subject"
            with self.temporary_permissions((rdef, {"add": ()})):
                user = self.vreg["etypes"].etype_class("CWUser")(req)
                autoform = self.vreg["forms"].select("edition", req, entity=user)
                self.assertFalse(
                    autoform.check_inlined_rdef_permissions(
                        rschema, role, tschema, ttype
                    )
                )
            # we actually don't care about the actual expression,
            # may_have_permission only checks the presence of such expressions
            expr = RRQLExpression("S use_email O")
            with self.temporary_permissions((rdef, {"add": (expr,)})):
                user = self.vreg["etypes"].etype_class("CWUser")(req)
                autoform = self.vreg["forms"].select("edition", req, entity=user)
                self.assertTrue(
                    autoform.check_inlined_rdef_permissions(
                        rschema, role, tschema, ttype
                    )
                )


class FormViewsCWTC(WebCWTC):
    def test_delete_conf_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWGroup X")
            self.view("deleteconf", rset, template=None, req=req).source

    def test_delete_conf_formview_composite(self):
        with self.admin_access.cnx() as cnx:
            d1 = cnx.create_entity("Directory", name="dtest1")
            d2 = cnx.create_entity("Directory", name="dtest2", parent=d1)
            d3 = cnx.create_entity("Directory", name="dtest3", parent=d2)
            d4 = cnx.create_entity("Directory", name="dtest4", parent=d1)
            for i in range(3):
                cnx.create_entity("Directory", name=f"child{i}", parent=d3)
            cnx.commit()

        class DirectoryDeleteView(DeleteConfFormView):
            __select__ = DeleteConfFormView.__select__ & is_instance("Directory")
            show_composite = True

        self.vreg["propertyvalues"]["navigation.page-size"] = 3
        with (
            self.admin_access.web_request() as req,
            self.temporary_appobjects(DirectoryDeleteView),
        ):
            rset = req.execute('Directory X WHERE X name "dtest1"')
            source = self.view(
                "deleteconf", rset, template=None, req=req
            ).source.decode("utf-8")
            # Show composite object at depth 1
            # Don't display "And more composite entities" since their are equal
            # to page size
            expected = (
                f"<li>"
                f'<a href="{BASE_URL}directory/{d1.eid}">dtest1</a>'
                f'<ul class="treeview"><li>'
                f'<a href="{BASE_URL}directory/{d4.eid}">dtest4</a>'
                f'</li><li class="last">'
                f'<a href="{BASE_URL}directory/{d2.eid}">dtest2</a>'
                f"</li></ul></li>"
            )
            self.assertIn(expected, source)

            # Page size is reached, show "And more composite entities"
            rset = req.execute('Directory X WHERE X name "dtest3"')
            source = self.view(
                "deleteconf", rset, template=None, req=req
            ).source.decode("utf-8")
            expected = (
                f"<li>"
                f'<a href="{BASE_URL}directory/%s">dtest3</a>'
                f'<ul class="treeview"><li>'
                f'<a href="{BASE_URL}directory/%s">child2</a>'
                f"</li><li>"
                f'<a href="{BASE_URL}directory/%s">child1</a>'
                f'</li><li class="last">And more composite entities</li>'
                f"</ul></li>"
            ) % (
                d3.eid,
                req.find("Directory", name="child2").one().eid,
                req.find("Directory", name="child1").one().eid,
            )
            self.assertIn(expected, source)

            # No composite entities
            rset = req.execute('Directory X WHERE X name "dtest4"')
            source = self.view(
                "deleteconf", rset, template=None, req=req
            ).source.decode("utf-8")
            expected = f'<li><a href="{BASE_URL}directory/{d4.eid}">' f"dtest4</a></li>"
            self.assertIn(expected, source)

    def test_automatic_edition_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X")
            self.view("edition", rset, row=0, template=None, req=req).source

    def test_automatic_edition_copyformview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X")
            self.view("copy", rset, row=0, template=None, req=req).source

    def test_automatic_creation_formview(self):
        with self.admin_access.web_request() as req:
            self.view("creation", None, etype="CWUser", template=None, req=req).source

    def test_automatic_muledit_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X")
            self.view("muledit", rset, template=None, req=req).source

    def test_automatic_reledit_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWUser X")
            self.view(
                "reledit", rset, row=0, rtype="login", template=None, req=req
            ).source

    def test_automatic_inline_edit_formview(self):
        with self.admin_access.web_request() as req:
            geid = req.execute("CWGroup X LIMIT 1")[0][0]
            rset = req.execute("CWUser X LIMIT 1")
            self.view(
                "inline-edition",
                rset,
                row=0,
                col=0,
                rtype="in_group",
                peid=geid,
                role="object",
                i18nctx="",
                pform=MOCKPFORM,
                template=None,
                req=req,
            ).source

    def test_automatic_inline_creation_formview(self):
        with self.admin_access.web_request() as req:
            geid = req.execute("CWGroup X LIMIT 1")[0][0]
            self.view(
                "inline-creation",
                None,
                etype="CWUser",
                rtype="in_group",
                peid=geid,
                petype="CWGroup",
                i18nctx="",
                role="object",
                pform=MOCKPFORM,
                template=None,
                req=req,
            )


MOCKPFORM = mock_object(form_previous_values={}, form_valerror=None)

if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import hashlib

from cubicweb_web.views import tableview
from cubicweb.server import Service
from cubicweb.predicates import has_related_entities, match_user_groups

from cubicweb_web.views.ajaxcontroller import ajaxfunc
from cubicweb_web.views.ibreadcrumbs import IBreadCrumbsAdapter
from cubicweb_web.views import xmlrss


xmlrss.RSSIconBox.visible = True


class TestService(Service):
    __regid__ = "test_service"
    __select__ = Service.__select__ & match_user_groups("managers")
    passed_here = []

    def call(self, msg):
        self.passed_here.append(msg)
        return "babar"


def _recursive_replace_stream_by_md5(tree):
    """Search for streams (i.e. object that have a 'read' method) in a tree
    (whose branches are lists or tuples), and substitute them by their md5 hash,
    leaving other leafs identical. A copy of the tree with only lists as
    branches is returned.
    """
    if not isinstance(tree, (list, tuple)):
        if hasattr(tree, "read"):
            return hashlib.md5(tree.read()).hexdigest()
        if hasattr(tree, "file"):
            return tree.filename, hashlib.md5(tree.file.read()).hexdigest()
        return tree
    else:
        return [_recursive_replace_stream_by_md5(value) for value in tree]


@ajaxfunc(output_type="json")
def fileupload(self):
    """Return a json copy of the web request form in which uploaded files
    are read and the received streams are replaced by their md5 hash.
    """
    try:
        result_dict = {}
        for key, value in self._cw.form.items():
            result_dict[key] = _recursive_replace_stream_by_md5(value)
        return result_dict
    except Exception as ex:
        import traceback as tb

        tb.print_exc(ex)


class FolderIBreadCrumbsAdapter(IBreadCrumbsAdapter):
    __select__ = IBreadCrumbsAdapter.__select__ & has_related_entities("filed_under")

    def parent_entity(self):
        return self.entity.filed_under[0]


class CustomRsetTableView(tableview.RsetTableView):
    __regid__ = "mytable"



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.utils"""
from urllib.parse import urljoin

from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.utils import HTMLHead


class HTMLHeadTC(WebCWTC):
    DATADIR_URL = urljoin(BASE_URL, "/data/")

    def htmlhead(self, datadir_url):
        with self.admin_access.web_request() as req:
            req.datadir_url = datadir_url
            head = HTMLHead(req)
            return head

    def test_concat_urls(self):
        base_url = self.DATADIR_URL
        head = self.htmlhead(base_url)
        urls = [base_url + "bob1.js", base_url + "bob2.js", base_url + "bob3.js"]
        result = head.concat_urls(urls)
        expected = urljoin(self.DATADIR_URL, "??bob1.js,bob2.js,bob3.js")
        self.assertEqual(result, expected)

    def test_group_urls(self):
        base_url = self.DATADIR_URL
        head = self.htmlhead(base_url)
        urls_spec = [
            (base_url + "bob0.js", None),
            (base_url + "bob1.js", None),
            ("http://ext.com/bob2.js", None),
            ("http://ext.com/bob3.js", None),
            (base_url + "bob4.css", "all"),
            (base_url + "bob5.css", "all"),
            (base_url + "bob6.css", "print"),
            (base_url + "bob7.css", "print"),
            (base_url + "bob8.css", ("all", "[if IE 8]")),
            (base_url + "bob9.css", ("print", "[if IE 8]")),
        ]
        result = head.group_urls(urls_spec)
        expected = [
            (base_url + "??bob0.js,bob1.js", None),
            ("http://ext.com/bob2.js", None),
            ("http://ext.com/bob3.js", None),
            (base_url + "??bob4.css,bob5.css", "all"),
            (base_url + "??bob6.css,bob7.css", "print"),
            (base_url + "bob8.css", ("all", "[if IE 8]")),
            (base_url + "bob9.css", ("print", "[if IE 8]")),
        ]
        self.assertEqual(list(result), expected)

    def test_getvalue_with_concat(self):
        self.config.global_set_option("concat-resources", True)
        base_url = self.DATADIR_URL
        head = self.htmlhead(base_url)
        head.add_js(base_url + "bob0.js")
        head.add_js(base_url + "bob1.js")
        head.add_js("http://ext.com/bob2.js")
        head.add_js("http://ext.com/bob3.js")
        head.add_css(base_url + "bob4.css")
        head.add_css(base_url + "bob5.css")
        head.add_css(base_url + "bob6.css", "print")
        head.add_css(base_url + "bob7.css", "print")
        result = head.getvalue()
        expected = f"""<head>
<link rel="stylesheet" type="text/css" media="all" href="{self.DATADIR_URL}??bob4.css,bob5.css"/>
<link rel="stylesheet" type="text/css" media="print" href="{self.DATADIR_URL}??bob6.css,bob7.css"/>
<script type="text/javascript" src="{self.DATADIR_URL}??bob0.js,bob1.js"></script>
<script type="text/javascript" src="http://ext.com/bob2.js"></script>
<script type="text/javascript" src="http://ext.com/bob3.js"></script>
</head>
"""  # noqa
        self.assertEqual(result, expected)

    def test_getvalue_without_concat(self):
        self.config.global_set_option("concat-resources", False)
        try:
            base_url = self.DATADIR_URL
            head = self.htmlhead(base_url)
            head.add_js(base_url + "bob0.js")
            head.add_js(base_url + "bob1.js")
            head.add_js("http://ext.com/bob2.js")
            head.add_js("http://ext.com/bob3.js")
            head.write_front(
                '<script type="text/javascript">console.log("FIRST SCRIPT ADDED HERE")</script>\n'
            )
            head.add_css(base_url + "bob4.css")
            head.add_css(base_url + "bob5.css")
            head.add_css(base_url + "bob6.css", "print")
            head.add_css(base_url + "bob7.css", "print")
            result = head.getvalue()
            expected = f"""<head>
<script type="text/javascript">console.log("FIRST SCRIPT ADDED HERE")</script>
<link rel="stylesheet" type="text/css" media="all" href="{self.DATADIR_URL}bob4.css"/>
<link rel="stylesheet" type="text/css" media="all" href="{self.DATADIR_URL}bob5.css"/>
<link rel="stylesheet" type="text/css" media="print" href="{self.DATADIR_URL}bob6.css"/>
<link rel="stylesheet" type="text/css" media="print" href="{self.DATADIR_URL}bob7.css"/>
<script type="text/javascript" src="{self.DATADIR_URL}bob0.js"></script>
<script type="text/javascript" src="{self.DATADIR_URL}bob1.js"></script>
<script type="text/javascript" src="http://ext.com/bob2.js"></script>
<script type="text/javascript" src="http://ext.com/bob3.js"></script>
</head>
"""
            self.assertEqual(result, expected)
        finally:
            self.config.global_set_option("concat-resources", True)

    def test_add_js_attribute(self):
        self.config.global_set_option("concat-resources", False)
        try:
            base_url = self.DATADIR_URL
            head = self.htmlhead(base_url)
            head.add_js(base_url + "bob0.js", script_attributes={"defer": True})
            head.add_js(
                "http://ext.com/bob2.js",
                script_attributes={"async": True, "integrity": "sha256-lala"},
            )
            result = head.getvalue()
            expected = f"""<head>
<script type="text/javascript" src="{self.DATADIR_URL}bob0.js" defer></script>
<script type="text/javascript" src="http://ext.com/bob2.js" async integrity="sha256-lala"></script>
</head>
"""
            self.assertEqual(result, expected)
        finally:
            self.config.global_set_option("concat-resources", True)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2016-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.server.sources import datafeed
from logilab.common import tempattr

from cubicweb_web.devtools.testlib import PyramidWebCWTC


class SynchronizeSourceTC(PyramidWebCWTC):
    def test_synchronize_view(self):
        with self.admin_access.web_request(vid="cw.source-sync") as req:

            class AParser(datafeed.DataFeedParser):
                __regid__ = "testparser"

                def process(self, url, raise_on_error=False):
                    pass

            with self.temporary_appobjects(AParser):
                source = req.create_entity(
                    "CWSource",
                    name="ext",
                    type="datafeed",
                    parser="cw.entityxml",
                    url="whatever",
                )
                req.cnx.commit()

            self.threads = 0

            def threaded_task(func):
                self.threads += 1

            with tempattr(req.cnx.repo, "threaded_task", threaded_task):
                path, args = self.expect_redirect_handle_request(
                    req, path=source.rest_path()
                )
                self.assertEqual(self.threads, 1)
                self.assertTrue(path.startswith("cwdataimport/"))


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common import tempattr

from cubicweb_web.devtools.testlib import WebCWTC, FakeRequest
from cubicweb_web.views.urlrewrite import (
    SimpleReqRewriter,
    SchemaBasedRewriter,
    rgx,
    rgx_action,
)


class UrlRewriteCWTC(WebCWTC):
    def test_auto_extend_rules(self):
        class Rewriter(SimpleReqRewriter):
            rules = [
                ("foo", dict(rql="Foo F")),
                ("/index", dict(vid="index2")),
            ]

        rules = []
        for pattern, values in Rewriter.rules:
            if hasattr(pattern, "pattern"):
                pattern = pattern.pattern
            rules.append((pattern, values))
        self.assertListEqual(
            rules,
            [
                ("foo", dict(rql="Foo F")),
                ("/index", dict(vid="index2")),
                ("/_", dict(vid="manage")),
                ("/_registry", dict(vid="registry")),
                ("/schema", dict(vid="schema")),
                ("/myprefs", dict(vid="propertiesform")),
                ("/siteconfig", dict(vid="systempropertiesform")),
                ("/siteinfo", dict(vid="siteinfo")),
                ("/manage", dict(vid="manage")),
                ("/notfound", dict(vid="404")),
                ("/error", dict(vid="error")),
                ("/processinfo", dict(vid="processinfo")),
                (
                    "/cwuser$",
                    {
                        "vid": "cw.users-and-groups-management",
                        "tab": "cw_users_management",
                    },
                ),
                (
                    "/cwgroup$",
                    {
                        "vid": "cw.users-and-groups-management",
                        "tab": "cw_groups_management",
                    },
                ),
                ("/cwsource$", {"vid": "cw.sources-management"}),
                (
                    "/schema/([^/]+?)/?$",
                    {"rql": r'Any X WHERE X is CWEType, X name "\1"', "vid": "primary"},
                ),
                ("/add/([^/]+?)/?$", dict(vid="creation", etype=r"\1")),
                ("/doc/images/(.+?)/?$", dict(fid="\\1", vid="wdocimages")),
                ("/doc/?$", dict(fid="main", vid="wdoc")),
                ("/doc/(.+?)/?$", dict(fid="\\1", vid="wdoc")),
                # now in SchemaBasedRewriter
                # ('/search/(.+)$', dict(rql=r'Any X WHERE X has_text "\1"')),
            ],
        )

    def test_no_extend_rules(self):
        class Rewriter(SimpleReqRewriter):
            ignore_baseclass_rules = True
            rules = [
                ("foo", dict(rql="Foo F")),
                ("/index", dict(vid="index2")),
            ]

        self.assertListEqual(
            Rewriter.rules,
            [
                ("foo", dict(rql="Foo F")),
                ("/index", dict(vid="index2")),
            ],
        )

    def test_basic_transformation(self):
        """test simple string-based rewrite"""
        req = FakeRequest()
        rewriter = SimpleReqRewriter(req)
        self.assertRaises(KeyError, rewriter.rewrite, req, "/view?vid=whatever")
        self.assertEqual(req.form, {})
        rewriter.rewrite(req, "/index")
        self.assertEqual(req.form, {"vid": "index"})

    def test_regexp_transformation(self):
        """test regexp-based rewrite"""
        req = FakeRequest()
        rewriter = SimpleReqRewriter(req)
        rewriter.rewrite(req, "/add/Task")
        self.assertEqual(req.form, {"vid": "creation", "etype": "Task"})
        req = FakeRequest()
        rewriter.rewrite(req, "/add/Task/")
        self.assertEqual(req.form, {"vid": "creation", "etype": "Task"})

    def test_inheritance(self):
        BaseTransition = self.vreg["etypes"].etype_class("BaseTransition")
        with self.admin_access.web_request() as req:
            x = req.create_entity("WorkflowTransition", name="test")
            ctrlid, rset = self.app.url_resolver.process(
                req, "basetransition/%s" % x.eid
            )
            self.assertEqual(ctrlid, "view")
            self.assertEqual(x.eid, rset[0][0])
            # cw_rest_attr_info is cached but clear_cache doesn't like cached class
            # method
            del BaseTransition._cw_rest_attr_info_cache_
            try:
                with tempattr(BaseTransition, "rest_attr", "name"):
                    ctrlid, rset = self.app.url_resolver.process(
                        req, "basetransition/%s" % x.name
                    )
                    self.assertEqual(ctrlid, "view")
                    self.assertEqual(x.eid, rset[0][0])
            finally:
                del BaseTransition._cw_rest_attr_info_cache_


class RgxActionRewriteCWTC(WebCWTC):
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            p1 = self.create_user(cnx, "user1")
            p1.cw_set(firstname="joe", surname="Dalton")
            p2 = self.create_user(cnx, "user2")
            p2.cw_set(firstname="jack", surname="Dalton")
            self.p1eid = p1.eid
            cnx.commit()

    def test_rgx_action_with_transforms(self):
        class TestSchemaBasedRewriter(SchemaBasedRewriter):
            rules = [
                (
                    rgx(r"/(?P<sn>\w+)/(?P<fn>\w+)"),
                    rgx_action(
                        r"Any X WHERE X surname %(sn)s, " "X firstname %(fn)s",
                        argsgroups=("sn", "fn"),
                        transforms={
                            "sn": str.capitalize,
                            "fn": str.lower,
                        },
                    ),
                ),
            ]

        with self.admin_access.web_request() as req:
            rewriter = TestSchemaBasedRewriter(req)
            _pmid, rset = rewriter.rewrite(req, "/DaLToN/JoE")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], self.p1eid)
            self.assertEqual(rset.description[0][0], "CWUser")

    def test_inheritance_precedence(self):
        RQL1 = "Any C WHERE C is CWEType"
        RQL2 = "Any C WHERE C is CWUser"

        class BaseRewriter(SchemaBasedRewriter):
            rules = [
                (
                    rgx("/collector(.*)"),
                    rgx_action(rql=RQL1, form=dict(vid="baseindex")),
                ),
            ]

        class Rewriter(BaseRewriter):
            rules = [
                (
                    rgx("/collector/something(/?)"),
                    rgx_action(rql=RQL2, form=dict(vid="index")),
                ),
            ]

        with self.admin_access.web_request() as req:
            rewriter = Rewriter(req)
            _pmid, rset = rewriter.rewrite(req, "/collector")
            self.assertEqual(rset.rql, RQL1)
            self.assertEqual(req.form, {"vid": "baseindex"})
            _pmid, rset = rewriter.rewrite(req, "/collector/something")
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {"vid": "index"})
            _pmid, rset = rewriter.rewrite(req, "/collector/something/")
            self.assertEqual(req.form, {"vid": "index"})
            self.assertEqual(rset.rql, RQL2)
            _pmid, rset = rewriter.rewrite(req, "/collector/somethingelse/")
            self.assertEqual(rset.rql, RQL1)
            self.assertEqual(req.form, {"vid": "baseindex"})

    def test_inheritance_precedence_same_rgx(self):
        RQL1 = "Any C WHERE C is CWEType"
        RQL2 = "Any C WHERE C is CWUser"

        class BaseRewriter(SchemaBasedRewriter):
            rules = [
                (
                    rgx("/collector(.*)"),
                    rgx_action(rql=RQL1, form=dict(vid="baseindex")),
                ),
            ]

        class Rewriter(BaseRewriter):
            rules = [
                (
                    rgx("/collector(.*)"),
                    rgx_action(rql=RQL2, form=dict(vid="index")),
                ),
            ]

        with self.admin_access.web_request() as req:
            rewriter = Rewriter(req)
            _pmid, rset = rewriter.rewrite(req, "/collector")
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {"vid": "index"})
            _pmid, rset = rewriter.rewrite(req, "/collector/something")
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {"vid": "index"})
            _pmid, rset = rewriter.rewrite(req, "/collector/something/")
            self.assertEqual(req.form, {"vid": "index"})
            self.assertEqual(rset.rql, RQL2)
            _pmid, rset = rewriter.rewrite(req, "/collector/somethingelse/")
            self.assertEqual(rset.rql, RQL2)
            self.assertEqual(req.form, {"vid": "index"})


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""base xml and rss views"""


from yams import xy

from cubicweb import _
from cubicweb.schema import VIRTUAL_RTYPES
from cubicweb_web.view import EntityView

try:
    import rdflib
except ImportError:
    rdflib = None

if rdflib is not None:
    RDF = rdflib.Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
    CW = rdflib.Namespace("http://ns.cubicweb.org/cubicweb/0.0/")
    from rdflib import Literal, URIRef, Namespace

    def urijoin(item):
        base, ext = item
        return URIRef(Namespace(base)[ext])

    SKIP_RTYPES = VIRTUAL_RTYPES | {"cwuri", "is", "is_instance_of"}

    class RDFView(EntityView):
        """rdf view for entities"""

        __regid__ = "rdf"
        title = _("rdf export")
        templatable = False
        binary = True
        format = "xml"  # or maybe pretty-xml ?
        content_type = "application/rdf+xml"

        def call(self):
            graph = rdflib.Graph()
            graph.bind("cw", CW)
            for prefix, xmlns in xy.XY.prefixes.items():
                graph.bind(prefix, rdflib.Namespace(xmlns))
            for i in range(self.cw_rset.rowcount):
                entity = self.cw_rset.complete_entity(i, 0)
                self.entity2graph(graph, entity)
            self.w(graph.serialize(format=self.format, encoding="utf-8"))

        def entity_call(self, entity):
            self.call()

        def entity2graph(self, graph, entity):
            cwuri = URIRef(entity.cwuri)
            add = graph.add
            add((cwuri, RDF.type, CW[entity.e_schema.type]))
            try:
                for item in xy.xeq(entity.e_schema.type):
                    add((cwuri, RDF.type, urijoin(item)))
            except xy.UnsupportedVocabulary:
                pass
            for rschema, eschemas, role in entity.e_schema.relation_definitions(
                "relation"
            ):
                rtype = rschema.type
                if rtype in SKIP_RTYPES or rtype.endswith("_permission"):
                    continue
                for eschema in eschemas:
                    if eschema.final:
                        try:
                            value = entity.cw_attr_cache[rtype]
                        except KeyError:
                            continue  # assuming rtype is Bytes
                        if value is not None:
                            add((cwuri, CW[rtype], Literal(value)))
                            try:
                                for item in xy.xeq(f"{entity.e_schema.type} {rtype}"):
                                    add((cwuri, urijoin(item[1]), Literal(value)))
                            except xy.UnsupportedVocabulary:
                                pass
                    else:
                        for related in entity.related(
                            rtype, role, entities=True, safe=True
                        ):
                            if role == "subject":
                                add((cwuri, CW[rtype], URIRef(related.cwuri)))
                                try:
                                    for item in xy.xeq(
                                        f"{entity.e_schema.type} {rtype}"
                                    ):
                                        add(
                                            (
                                                cwuri,
                                                urijoin(item[1]),
                                                URIRef(related.cwuri),
                                            )
                                        )
                                except xy.UnsupportedVocabulary:
                                    pass
                            else:
                                add((URIRef(related.cwuri), CW[rtype], cwuri))

    class RDFN3View(RDFView):
        __regid__ = "rdf.n3"
        format = "n3"
        content_type = "text/n3"  # see https://www.w3.org/TeamSubmission/n3/#mimetype

    class RDFTurtleView(RDFView):
        __regid__ = "rdf.turtle"
        format = "turtle"
        content_type = "text/turtle"  # see https://www.w3.org/TR/turtle/#sec-mediaReg

    class RDFntriplesView(RDFView):
        __regid__ = "rdf.nt"
        format = "nt"
        content_type = "application/n-triples"  # see https://www.w3.org/TR/n-triples/#n-triples-mediatype

    class RDFtrigView(RDFView):
        __regid__ = "rdf.trig"
        format = "trig"
        content_type = "application/trig"  # see https://www.w3.org/TR/trig/#sec-mime



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Primary view for bookmarks + user's bookmarks box"""


from cubicweb import _

from cubicweb.predicates import is_instance, one_line_rset
from cubicweb_web import action, component, htmlwidgets, formwidgets as fw
from cubicweb_web.views import uicfg, primary
from cubicweb_web.views.ajaxcontroller import ajaxfunc

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_subject_of(("*", "bookmarked_by", "*"), False)
_abaa.tag_object_of(("*", "bookmarked_by", "*"), False)

_afs = uicfg.autoform_section
_afs.tag_object_of(("*", "bookmarked_by", "CWUser"), "main", "metadata")
_afs.tag_attribute(("Bookmark", "path"), "main", "attributes")
_afs.tag_attribute(("Bookmark", "path"), "muledit", "attributes")

_affk = uicfg.autoform_field_kwargs
_affk.tag_attribute(("Bookmark", "path"), {"widget": fw.EditableURLWidget})


class FollowAction(action.Action):
    __regid__ = "follow"
    __select__ = one_line_rset() & is_instance("Bookmark")

    title = _("follow")
    category = "mainactions"

    def url(self):
        return self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0).actual_url()


class BookmarkPrimaryView(primary.PrimaryView):
    __select__ = is_instance("Bookmark")

    def cell_call(self, row, col):
        """the primary view for bookmark entity"""
        entity = self.cw_rset.complete_entity(row, col)
        self.w("&#160;")
        self.w("<span class='title'><b>")
        self.w("%s : %s", self._cw._("Bookmark"), entity.title)
        self.w("</b></span>")
        self.w('<br/><br/><div class="content"><a href="%s">', entity.actual_url())
        self.w("</a>")
        self.w(
            "<p>%s%s</p>",
            self._cw._("Used by:"),
            ", ".join(u.name() for u in entity.bookmarked_by),
        )
        self.w("</div>")


class BookmarksBox(component.CtxComponent):
    """display a box containing all user's bookmarks"""

    __regid__ = "bookmarks_box"

    title = _("bookmarks")
    order = 40
    rql = (
        "Any B,T,P ORDERBY lower(T) "
        "WHERE B is Bookmark,B title T, B path P, B bookmarked_by U, "
        "U eid %(x)s"
    )

    def init_rendering(self):
        ueid = self._cw.user.eid
        self.bookmarks_rset = self._cw.execute(self.rql, {"x": ueid})
        rschema = self._cw.vreg.schema.relation_schema_for("bookmarked_by")
        eschema = self._cw.vreg.schema.entity_schema_for("Bookmark")
        self.can_delete = rschema.has_perm(self._cw, "delete", toeid=ueid)
        self.can_edit = eschema.has_perm(self._cw, "add") and rschema.has_perm(
            self._cw, "add", toeid=ueid
        )
        if not self.bookmarks_rset and not self.can_edit:
            raise component.EmptyComponent()
        self.items = []

    def render_body(self, w):
        ueid = self._cw.user.eid
        req = self._cw
        if self.can_delete:
            req.add_js("cubicweb.ajax.js")
        for bookmark in self.bookmarks_rset.entities():
            label = self.link(bookmark.title, bookmark.action_url())
            if self.can_delete:
                dlink = (
                    '[<a class="action" href="javascript:removeBookmark(%s)" title="%s">-</a>]'
                    % (bookmark.eid, req._("delete this bookmark"))
                )
                label = f"<div>{dlink} {label}</div>"
            self.append(label)
        if self.can_edit:
            menu = htmlwidgets.BoxMenu(req._("manage bookmarks"))
            linkto = "bookmarked_by:%s:subject" % ueid
            # use a relative path so that we can move the instance without
            # loosing bookmarks
            path = req.relative_path()
            # XXX if vtitle specified in params, extract it and use it as
            # default value for bookmark's title
            url = (
                req.vreg["etypes"]
                .etype_class("Bookmark")
                .cw_create_url(req, __linkto=linkto, path=path)
            )
            menu.append(self.link(req._("bookmark this page"), url))
            if self.bookmarks_rset:
                if req.user.is_in_group("managers"):
                    bookmarksrql = (
                        "Any B ORDERBY T WHERE B is Bookmark, B title T, "
                        "B bookmarked_by U, U eid %s" % ueid
                    )
                    erset = self.bookmarks_rset
                else:
                    # we can't edit shared bookmarks we don't own
                    bookmarksrql = (
                        "Any B ORDERBY T WHERE B is Bookmark, B title T, "
                        "B bookmarked_by U, B owned_by U, U eid %(x)s"
                    )
                    erset = req.execute(bookmarksrql, {"x": ueid}, build_descr=False)
                    bookmarksrql %= {"x": ueid}
                if erset:
                    url = req.build_url(vid="muledit", rql=bookmarksrql)
                    menu.append(self.link(req._("edit bookmarks"), url))
            url = req.user.absolute_url(
                vid="xaddrelation", rtype="bookmarked_by", target="subject"
            )
            menu.append(self.link(req._("pick existing bookmarks"), url))
            self.append(menu)
        self.render_items(w)


@ajaxfunc
def delete_bookmark(self, beid):
    rql = "DELETE B bookmarked_by U WHERE B eid %(b)s, U eid %(u)s"
    self._cw.execute(rql, {"b": int(beid), "u": self._cw.user.eid})



from io import BytesIO
from urllib.parse import urljoin

import webtest
import pyramid.request

from cubicweb.devtools import BASE_URL

from cubicweb_web.bwcompat import CubicWebPyramidRequest
from cubicweb_web.devtools.testlib import PyramidWebCWTC


class WSGIAppTest(PyramidWebCWTC):
    def make_request(self, path, environ=None, **kw):
        r = webtest.app.TestRequest.blank(urljoin(BASE_URL, path), environ, **kw)

        request = pyramid.request.Request(r.environ)
        request.registry = self.pyr_registry

        return request

    def test_content_type(self):
        req = CubicWebPyramidRequest(
            self.make_request(BASE_URL, {"CONTENT_TYPE": "text/plain"})
        )

        self.assertEqual("text/plain", req.get_header("Content-Type"))

    def test_content_body(self):
        req = CubicWebPyramidRequest(
            self.make_request(
                BASE_URL,
                {
                    "CONTENT_LENGTH": 12,
                    "CONTENT_TYPE": "text/plain",
                    "wsgi.input": BytesIO(b"some content"),
                },
            )
        )

        self.assertEqual(b"some content", req.content.read())

    def test_big_content(self):
        content = b"x" * 100001

        req = CubicWebPyramidRequest(
            self.make_request(
                BASE_URL,
                {
                    "CONTENT_LENGTH": len(content),
                    "CONTENT_TYPE": "text/plain",
                    "wsgi.input": BytesIO(content),
                },
            )
        )

        self.assertEqual(content, req.content.read())

    def test_post(self):
        self.webapp.post(
            BASE_URL, params={"__login": self.admlogin, "__password": self.admpassword}
        )

    def test_get_multiple_variables(self):
        req = CubicWebPyramidRequest(self.make_request("/?arg=1&arg=2"))

        self.assertEqual(["1", "2"], req.form["arg"])

    def test_post_multiple_variables(self):
        req = CubicWebPyramidRequest(self.make_request("/", POST="arg=1&arg=2"))

        self.assertEqual(["1", "2"], req.form["arg"])

    def test_post_files(self):
        content_type, params = self.webapp.encode_multipart(
            (), (("filefield", "aname", b"acontent"),)
        )
        req = CubicWebPyramidRequest(
            self.make_request("/", POST=params, content_type=content_type)
        )
        self.assertIn("filefield", req.form)
        fieldvalue = req.form["filefield"]
        self.assertEqual("aname", fieldvalue[0])
        self.assertEqual(b"acontent", fieldvalue[1].read())

    def test_post_unicode_urlencoded(self):
        params = "arg=%C3%A9"
        req = CubicWebPyramidRequest(
            self.make_request(
                "/", POST=params, content_type="application/x-www-form-urlencoded"
            )
        )
        self.assertEqual("é", req.form["arg"])

    def test_404(self):
        r = self.webapp.get(urljoin(BASE_URL, "unexisting/"), status=404)
        self.assertNotIn("error occurred", r.text)
        self.assertIn("resource does not exist", r.text)


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""CubicWeb web client application object"""


import contextlib
import http.client as http_client
import json
import sys
from contextlib import contextmanager
from time import process_time, time

from rql import BadRQLQuery

from cubicweb import (
    CW_EVENT_MANAGER,
    ValidationError,
    Unauthorized,
    Forbidden,
    AuthenticationError,
    NoSelectableObject,
)
from cubicweb import set_log_methods
from cubicweb.repoapi import anonymous_cnx
from cubicweb_web import (
    LOGGER,
    DirectResponse,
    Redirect,
    NotFound,
    LogOut,
    RemoteCallFailed,
    InvalidSession,
    RequestError,
    PublishException,
)

# make session manager available through a global variable so the debug view can
# print information about web session
SESSION_MANAGER = None


@contextmanager
def anonymized_request(req):
    from cubicweb_web.views.authentication import Session

    orig_cnx = req.cnx
    anon_cnx = anonymous_cnx(orig_cnx.repo)
    try:
        with anon_cnx:
            # web request expect a session attribute on cnx referencing the web session
            anon_cnx.session = Session(orig_cnx.repo, anon_cnx.user)
            req.set_cnx(anon_cnx)
            yield req
    finally:
        req.set_cnx(orig_cnx)


class CookieSessionHandler:
    """a session handler using a cookie to store the session identifier"""

    def __init__(self, appli):
        self.repo = appli.repo
        self.vreg = appli.vreg
        self.session_manager = self.vreg["sessions"].select(
            "sessionmanager", repo=self.repo
        )
        global SESSION_MANAGER
        SESSION_MANAGER = self.session_manager
        if self.vreg.config.mode != "test":
            # don't try to reset session manager during test, this leads to
            # weird failures when running multiple tests
            CW_EVENT_MANAGER.bind("after-registry-reload", self.reset_session_manager)

    def reset_session_manager(self):
        data = self.session_manager.dump_data()
        self.session_manager = self.vreg["sessions"].select(
            "sessionmanager", repo=self.repo
        )
        self.session_manager.restore_data(data)
        global SESSION_MANAGER
        SESSION_MANAGER = self.session_manager

    @property
    def clean_sessions_interval(self):
        return self.session_manager.clean_sessions_interval

    def clean_sessions(self):
        """cleanup sessions which has not been unused since a given amount of
        time
        """
        self.session_manager.clean_sessions()

    def session_cookie(self, req):
        """return a string giving the name of the cookie used to store the
        session identifier.
        """
        return "__%s_session" % self.vreg.config.appid

    def get_session(self, req):
        """Return a session object corresponding to credentials held by the req

        Session id is searched from :
        - # form variable
        - cookie

        If no session id is found, try opening a new session with credentials
        found in the request.

        Raises AuthenticationError if no session can be found or created.
        """
        cookie = req.get_cookie()
        sessioncookie = self.session_cookie(req)
        try:
            sessionid = str(cookie[sessioncookie].value)
            session = self.get_session_by_id(req, sessionid)
        except (KeyError, InvalidSession):  # no valid session cookie
            session = self.open_session(req)
        return session

    def get_session_by_id(self, req, sessionid):
        session = self.session_manager.get_session(req, sessionid)
        session.mtime = time()
        return session

    def open_session(self, req):
        session = self.session_manager.open_session(req)
        sessioncookie = self.session_cookie(req)
        secure = req.base_url().startswith("https://")
        req.set_cookie(
            sessioncookie, session.sessionid, maxage=None, secure=secure, httponly=True
        )
        if not session.anonymous_session:
            self.session_manager.postlogin(req, session)
        return session

    def logout(self, req, goto_url):
        """logout from the instance by cleaning the session and raising
        `AuthenticationError`
        """
        self.session_manager.close_session(req.session)
        req.remove_cookie(self.session_cookie(req))
        raise LogOut(url=goto_url)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


class CubicWebPublisher:
    """the publisher is a singleton hold by the web frontend, and is responsible
    to publish HTTP request.

    The http server will call its main entry point ``application.handle_request``.

    .. automethod:: cubicweb_web.application.CubicWebPublisher.main_handle_request

    You have to provide both a repository and web-server config at
    initialization. In all in one instance both config will be the same.
    """

    def __init__(self, repo, config, session_handler_fact=CookieSessionHandler):
        self.info("starting web instance from %s", config.apphome)
        self.repo = repo
        self.vreg = repo.vreg
        # get instance's schema
        if not self.vreg.initialized:
            config.init_cubes(self.repo.get_cubes())
            self.vreg.init_properties(self.repo.properties())
            self.vreg.set_schema(self.repo.get_schema())
        # set the correct publish method
        query_log_file = config.get("query-log-file")
        if query_log_file:
            from threading import Lock

            self._query_log = open(query_log_file, "a")
            self.handle_request = self.log_handle_request
            self._logfile_lock = Lock()
        else:
            self._query_log = None
            self.handle_request = self.main_handle_request
        # instantiate session and url resolving helpers
        self.session_handler = session_handler_fact(self)
        self.set_urlresolver()
        CW_EVENT_MANAGER.bind("after-registry-reload", self.set_urlresolver)

    def set_urlresolver(self):
        self.url_resolver = self.vreg["components"].select(
            "urlpublisher", vreg=self.vreg
        )

    def get_session(self, req):
        """Return a session object corresponding to credentials held by the req

        May raise AuthenticationError.
        """
        return self.session_handler.get_session(req)

    # publish methods #########################################################

    def log_handle_request(self, req):
        """wrapper around _publish to log all queries executed for a given
        accessed path
        """

        def wrap_set_cnx(func):
            def wrap_execute(cnx):
                orig_execute = cnx.execute

                def execute(rql, kwargs=None, build_descr=True):
                    tstart, cstart = time(), process_time()
                    rset = orig_execute(rql, kwargs, build_descr=build_descr)
                    cnx.executed_queries.append(
                        (rql, kwargs, time() - tstart, process_time() - cstart)
                    )
                    return rset

                return execute

            def set_cnx(cnx):
                func(cnx)
                cnx.execute = wrap_execute(cnx)
                cnx.executed_queries = []

            return set_cnx

        req.set_cnx = wrap_set_cnx(req.set_cnx)
        tstart, cstart = time(), process_time()
        try:
            return self.main_handle_request(req)
        finally:
            cnx = req.cnx
            if cnx and cnx.executed_queries:
                with self._logfile_lock:
                    tend, cend = time(), process_time()
                    try:
                        result = ["\n" + "*" * 80]
                        result.append(
                            "%s -- (%.3f sec, %.3f CPU sec)"
                            % (req.url(), tend - tstart, cend - cstart)
                        )
                        result += [
                            "%s %s -- (%.3f sec, %.3f CPU sec)" % q
                            for q in cnx.executed_queries
                        ]
                        cnx.executed_queries = []
                        self._query_log.write("\n".join(result))
                        self._query_log.flush()
                    except Exception:
                        self.exception("error while logging queries")

    def main_handle_request(self, req):
        """Process an HTTP request `req`

        :type req: `web.Request`
        :param req: the request object

        It returns the content of the http response. HTTP header and status are
        set on the Request object.
        """
        if req.authmode == "http":
            # activate realm-based auth
            realm = self.vreg.config["realm"]
            req.set_header("WWW-Authenticate", [("Basic", {"realm": realm})], raw=False)
        content = b""
        try:
            try:
                session = self.get_session(req)
                cnx = session.new_cnx()
                with cnx:  # may need an open connection to access to e.g. properties
                    req.set_cnx(cnx)
                cnx._open = None  # XXX needed to reuse it a few line later :'(
            except AuthenticationError:
                # Keep the dummy session set at initialisation.  such session will work to some
                # extend but raise an AuthenticationError on any database access.
                # XXX We want to clean up this approach in the future. But several cubes like
                # registration or forgotten password rely on this principle.
                @contextlib.contextmanager
                def dummy():
                    yield

                cnx = dummy()
            # nested try to allow LogOut to delegate logic to AuthenticationError
            # handler
            try:
                # Try to generate the actual request content
                with cnx:
                    content = self.core_handle(req)
            # Handle user log-out
            except LogOut as ex:
                # When authentification is handled by cookie the code that
                # raised LogOut must has invalidated the cookie. We can just
                # reload the original url without authentification
                if self.vreg.config["auth-mode"] == "cookie" and ex.url:
                    req.headers_out.setHeader("location", str(ex.url))
                if ex.status is not None:
                    req.status_out = http_client.SEE_OTHER
                # When the authentification is handled by http we must
                # explicitly ask for authentification to flush current http
                # authentification information
                else:
                    # Render "logged out" content.
                    # assignement to ``content`` prevent standard
                    # AuthenticationError code to overwrite it.
                    content = self.loggedout_content(req)
                    # let the explicitly reset http credential
                    raise AuthenticationError()
        except Redirect as ex:
            # authentication needs redirection (eg openid)
            content = self.redirect_handler(req, ex)
        # Wrong, absent or Reseted credential
        except AuthenticationError:
            # We assume here that in http auth mode the user *May* provide
            # Authentification Credential if asked kindly.
            if self.vreg.config["auth-mode"] == "http":
                req.status_out = http_client.UNAUTHORIZED
            # In the other case (coky auth) we assume that there is no way
            # for the user to provide them...
            # XXX But WHY ?
            else:
                req.status_out = http_client.FORBIDDEN
            # If previous error handling already generated a custom content
            # do not overwrite it. This is used by LogOut Except
            # XXX ensure we don't actually serve content
            if not content:
                content = self.need_login_content(req)
        assert isinstance(content, bytes)
        return content

    def core_handle(self, req):
        """method called by the main publisher to process <req> relative path

        should return a string containing the resulting page or raise a
        `NotFound` exception

        :type req: `web.Request`
        :param req: the request object

        :rtype: str
        :return: the result of the pusblished url
        """
        path = req.relative_path(False)
        # don't log form values they may contains sensitive information
        self.debug(
            'publish "%s" (%s, form params: %s)',
            path,
            req.session.sessionid,
            list(req.form),
        )
        # remove user callbacks on a new request (except for json controllers
        # to avoid callbacks being unregistered before they could be called)
        tstart = process_time()
        commited = False
        try:
            # standard processing of the request
            try:
                # The process function update the req according rewrites rules.
                # It stores in req._rql_rewritten if the rql comes from the endUser.
                ctrlid, rset = self.url_resolver.process(req, path)
                try:
                    controller = self.vreg["controllers"].select(
                        ctrlid, req, appli=self
                    )
                except NoSelectableObject:
                    raise Unauthorized(req._("not authorized"))
                req.update_search_state()
                # In case of rewritten rql, controller.publish directly raises
                # notFound if the rset is empty.
                result = controller.publish(rset=rset)
            except Redirect as ex:
                # Redirect may be raised by edit controller when everything went
                # fine, so attempt to commit
                result = self.redirect_handler(req, ex)
            if req.cnx:
                txuuid = req.cnx.commit()
                commited = True
                if txuuid is not None:
                    req.data["last_undoable_transaction"] = txuuid
        # error case
        except NotFound as ex:
            result = self.notfound_content(req)
            req.status_out = ex.status
        except ValidationError as ex:
            result = self.validation_error_handler(req, ex)
        except RemoteCallFailed as ex:
            result = self.ajax_error_handler(req, ex)
        except Unauthorized as ex:
            req.data["errmsg"] = req._(
                "You're not authorized to access this page. "
                "If you think you should, please contact the site administrator."
            )
            req.status_out = http_client.FORBIDDEN
            result = self.error_handler(req, ex, tb=False)
        except Forbidden as ex:
            req.data["errmsg"] = req._(
                "This action is forbidden. "
                "If you think it should be allowed, please contact the site administrator."
            )
            req.status_out = http_client.FORBIDDEN
            result = self.error_handler(req, ex, tb=False)
        except (BadRQLQuery, RequestError) as ex:
            result = self.error_handler(req, ex, tb=False)
        # pass through exception
        except DirectResponse:
            if req.cnx:
                req.cnx.commit()
            raise
        except (AuthenticationError, LogOut):
            # the rollback is handled in the finally
            raise
        # Last defense line
        except BaseException as ex:
            req.status_out = http_client.INTERNAL_SERVER_ERROR
            result = self.error_handler(req, ex, tb=True)
        finally:
            if req.cnx and not commited:
                try:
                    req.cnx.rollback()
                except Exception:
                    pass  # ignore rollback error at this point
        self.add_undo_link_to_msg(req)
        self.debug("query %s executed in %s sec", path, process_time() - tstart)
        return result

    # Error handlers

    def redirect_handler(self, req, ex):
        """handle redirect
        - comply to ex status
        - set header field
        - return empty content
        """
        self.debug("redirecting to %s", str(ex.location))
        req.headers_out.setHeader("location", str(ex.location))
        assert 300 <= ex.status < 400
        req.status_out = ex.status
        return b""

    def validation_error_handler(self, req, ex):
        ex.translate(req._)  # translate messages using ui language
        if "__errorurl" in req.form:
            forminfo = {
                "error": ex,
                "values": req.form,
                "eidmap": req.data.get("eidmap", {}),
            }
            req.session.data[req.form["__errorurl"]] = forminfo
            # XXX form session key / __error_url should be differentiated:
            # session key is 'url + #<form dom id', though we usually don't want
            # the browser to move to the form since it hides the global
            # messages.
            location = req.form["__errorurl"].rsplit("#", 1)[0]
            req.headers_out.setHeader("location", str(location))
            req.status_out = http_client.SEE_OTHER
            return b""
        req.status_out = http_client.CONFLICT
        return self.error_handler(req, ex, tb=False)

    def error_handler(self, req, ex, tb=False):
        excinfo = sys.exc_info()
        if tb:
            self.exception(repr(ex))
        req.set_header("Cache-Control", "no-cache")
        req.remove_header("Etag")
        req.remove_header("Content-disposition")
        req.reset_message()
        req.reset_headers()
        if req.ajax_request:
            return self.ajax_error_handler(req, ex)
        try:
            req.data["ex"] = ex
            if tb:
                req.data["excinfo"] = excinfo
            errview = self.vreg["views"].select("error", req)
            template = self.main_template_id(req)
            content = self.vreg["views"].main_template(req, template, view=errview)
        except Exception:
            content = self.vreg["views"].main_template(req, "error-template")
        if isinstance(ex, PublishException) and ex.status is not None:
            req.status_out = ex.status
        return content

    def add_undo_link_to_msg(self, req):
        txuuid = req.data.get("last_undoable_transaction")
        if txuuid is not None:
            msg = '<span class="undo">[<a href="{}">{}</a>]</span>'.format(
                req.build_url("undo", txuuid=txuuid),
                req._("undo"),
            )
            req.append_to_redirect_message(msg)

    def ajax_error_handler(self, req, ex):
        req.set_header("content-type", "application/json")
        status = http_client.INTERNAL_SERVER_ERROR
        if isinstance(ex, PublishException) and ex.status is not None:
            status = ex.status
        if req.status_out < 400:
            # don't overwrite it if it's already set
            req.status_out = status
        json_dumper = getattr(ex, "dumps", lambda: json.dumps({"reason": str(ex)}))
        return json_dumper().encode("utf-8")

    # special case handling

    def need_login_content(self, req):
        return self.vreg["views"].main_template(req, "login")

    def loggedout_content(self, req):
        return self.vreg["views"].main_template(req, "loggedout")

    def notfound_content(self, req):
        req.form["vid"] = "404"
        view = self.vreg["views"].select("404", req)
        template = self.main_template_id(req)
        return self.vreg["views"].main_template(req, template, view=view)

    # template stuff

    def main_template_id(self, req):
        template = req.form.get("__template", req.property_value("ui.main-template"))
        if template not in self.vreg["views"]:
            template = "main-template"
        return template

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


set_log_methods(CubicWebPublisher, LOGGER)
set_log_methods(CookieSessionHandler, LOGGER)
